#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.timetable.schedule
"""
import doctest
import pytz
import unittest
from pprint import pprint
from datetime import date, time

from schooltool.timetable.schedule import date_timespan
from schooltool.timetable.schedule import iterMeetingsInTimezone
from schooltool.timetable.tests import ScheduleStub


def test_date_timespan():
    """Tests for date_timespan.

    Let's look a simple UTC date

       >>> date_timespan(date(2011, 10, 30), pytz.UTC)
       (datetime.datetime(2011, 10, 30, 0, 0, tzinfo=<UTC>),
        datetime.datetime(2011, 10, 30, 23, 59, 59, 999999, tzinfo=<UTC>))

    Let's look at date with daylight saving transition.

       >>> span = date_timespan(
       ...     date(2011, 10, 30), pytz.timezone('Europe/Vilnius'))

       >>> span
       (datetime.datetime(2011, 10, 30, 0, 0,
            tzinfo=<DstTzInfo 'Europe/Vilnius' EEST+3:00:00 DST>),
       datetime.datetime(2011, 10, 30, 23, 59, 59, 999999,
            tzinfo=<DstTzInfo 'Europe/Vilnius' EET+2:00:00 STD>))

       >>> print span[1] - span[0]
       1 day, 0:59:59.999999

    And another transition in spring.

       >>> span = date_timespan(
       ...     date(2011, 3, 27), pytz.timezone('Europe/Vilnius'))

       >>> span
       (datetime.datetime(2011, 3, 27, 0, 0,
             tzinfo=<DstTzInfo 'Europe/Vilnius' EET+2:00:00 STD>),
        datetime.datetime(2011, 3, 27, 23, 59, 59, 999999,
             tzinfo=<DstTzInfo 'Europe/Vilnius' EEST+3:00:00 DST>))

       >>> print span[1] - span[0]
       22:59:59.999999

    Note, this implementation does not fully cover time shifts that happen at
    midnight.  Consider 1915 in Warsaw:

       >>> span1 = date_timespan(
       ...     date(1915, 8, 4), pytz.timezone('Europe/Warsaw'))

       >>> span1
       (datetime.datetime(1915, 8, 4, 0, 0,
            tzinfo=<DstTzInfo 'Europe/Warsaw' WMT+1:24:00 STD>),
        datetime.datetime(1915, 8, 4, 23, 59, 59, 999999,
            tzinfo=<DstTzInfo 'Europe/Warsaw' WMT+1:24:00 STD>))

       >>> span2 = date_timespan(
       ...     date(1915, 8, 5), pytz.timezone('Europe/Warsaw'))

       >>> span2
       (datetime.datetime(1915, 8, 5, 0, 0,
            tzinfo=<DstTzInfo 'Europe/Warsaw' CET+1:00:00 STD>),
        datetime.datetime(1915, 8, 5, 23, 59, 59, 999999,
            tzinfo=<DstTzInfo 'Europe/Warsaw' CET+1:00:00 STD>))

    But there was a 24 minute time shift when they switched from
    Warsaw Mean Time to Central European Time:

       >>> print span1[1] - span1[0]
       23:59:59.999999

       >>> print span2[1] - span2[0]
       23:59:59.999999

       >>> print (span1[1] - span1[0]) + (span2[1] - span2[0])
       1 day, 23:59:59.999998

       >>> print span2[1] - span1[0]
       2 days, 0:23:59.999999

    """


def test_iterMeetingsInTimezone():
    """Tests for iterMeetingsInTimezone.

    Let's define a simple schedule in UTC with meetings every hour.

        >>> schedule = ScheduleStub(
        ...      timezone='UTC',
        ...      first=date(2011, 10, 28),
        ...      last=date(2011, 11, 01))

        >>> schedule.meeting_times = tuple(
        ...     [time(h, 0) for h in range(24)])

        >>> pprint(list(schedule.iterMeetings(schedule.first, schedule.last)))
        [<Meeting on 2011-10-28 00:00 UTC>,
         <Meeting on 2011-10-28 01:00 UTC>,
         <Meeting on 2011-10-28 02:00 UTC>,
         <Meeting on 2011-10-28 03:00 UTC>,
         ...
         <Meeting on 2011-11-01 22:00 UTC>,
         <Meeting on 2011-11-01 23:00 UTC>]

    Let's iterate over meetings that occur at 2011 10 29 in Vilnius:

        >>> meetings_29_vilnius = list(iterMeetingsInTimezone(
        ...    schedule, 'Europe/Vilnius', date(2011, 10, 29)))

    Note that there are 24 meetings, as expected:

        >>> len(meetings_29_vilnius)
        24

        >>> pprint(meetings_29_vilnius)
        [<Meeting on 2011-10-28 21:00 UTC>,
         <Meeting on 2011-10-28 22:00 UTC>,
         <Meeting on 2011-10-28 23:00 UTC>,
         ...
         <Meeting on 2011-10-29 18:00 UTC>,
         <Meeting on 2011-10-29 19:00 UTC>,
         <Meeting on 2011-10-29 20:00 UTC>]

    But at Oct 30 there's a daylight saving switch in Vilnius:

        >>> meetings_30_vilnius = list(iterMeetingsInTimezone(
        ...    schedule, 'Europe/Vilnius', date(2011, 10, 30)))

        >>> len(meetings_30_vilnius)
        25

        >>> pprint(meetings_30_vilnius)
        [<Meeting on 2011-10-29 21:00 UTC>,
         <Meeting on 2011-10-29 22:00 UTC>,
         <Meeting on 2011-10-29 23:00 UTC>,
         ...
         <Meeting on 2011-10-30 17:00 UTC>,
         <Meeting on 2011-10-30 18:00 UTC>,
         <Meeting on 2011-10-30 19:00 UTC>,
         <Meeting on 2011-10-30 20:00 UTC>,
         <Meeting on 2011-10-30 21:00 UTC>]

        >>> len(list(iterMeetingsInTimezone(
        ...     schedule, 'Europe/Vilnius',
        ...     date(2011, 10, 29), date(2011, 10, 30))))
        49

    Same happens with sprint dst shift.

        >>> schedule.first = date(2011, 3, 25)
        >>> schedule.last = date(2011, 3, 28)

    Winter time:

        >>> pprint(list(iterMeetingsInTimezone(
        ...     schedule, 'Europe/Vilnius',
        ...     date(2011, 3, 26))))
        [<Meeting on 2011-03-25 22:00 UTC>,
         ...
         <Meeting on 2011-03-26 21:00 UTC>]

    Observe the shift:

        >>> pprint(list(iterMeetingsInTimezone(
        ...     schedule, 'Europe/Vilnius',
        ...     date(2011, 3, 27))))
        [<Meeting on 2011-03-26 22:00 UTC>,
         ...
         <Meeting on 2011-03-27 20:00 UTC>]

        >>> len(list(iterMeetingsInTimezone(
        ...     schedule, 'Europe/Vilnius',
        ...     date(2011, 3, 27))))
        23

    Back to summer time:

        >>> pprint(list(iterMeetingsInTimezone(
        ...     schedule, 'Europe/Vilnius',
        ...     date(2011, 3, 28))))
        [<Meeting on 2011-03-27 21:00 UTC>,
         ...
         <Meeting on 2011-03-28 20:00 UTC>]

    """


def test_iterMeetingsInTimezone_localized_schedule():
    """Tests for iterMeetingsInTimezone.

    Let's define a simple schedule in Europe/Vilnius.

        >>> schedule = ScheduleStub(
        ...      timezone='Europe/Vilnius',
        ...      first=date(2011, 3, 12),
        ...      last=date(2011, 3, 28))

    We can see that dst shift occurs in this schedule:

        >>> pprint(list(schedule.iterMeetings(schedule.first, schedule.last)))
        [<Meeting on 2011-03-12 00:05 EET>,
         <Meeting on 2011-03-12 05:00 EET>,
         <Meeting on 2011-03-12 23:55 EET>,
         <Meeting on 2011-03-13 00:05 EET>,
         ...
         <Meeting on 2011-03-27 00:05 EET>,
         <Meeting on 2011-03-27 05:00 EEST>,
         <Meeting on 2011-03-27 23:55 EEST>,
         <Meeting on 2011-03-28 00:05 EEST>,
         <Meeting on 2011-03-28 05:00 EEST>,
         <Meeting on 2011-03-28 23:55 EEST>]

    Let's set up meetings every hour:

        >>> schedule.meeting_times = tuple(
        ...     [time(h, 0) for h in range(24)])

    We have 24 meetings per day:

        >>> pprint(list(iterMeetingsInTimezone(
        ...     schedule, 'UTC', date(2011, 3, 26))))
        [<Meeting on 2011-03-26 02:00 EET>,
         <Meeting on 2011-03-26 03:00 EET>,
         ...
         <Meeting on 2011-03-27 00:00 EET>,
         <Meeting on 2011-03-27 01:00 EET>]

        >>> len(list(iterMeetingsInTimezone(
        ...     schedule, 'UTC', date(2011, 3, 26))))
        24

    But because of spring dst shift, we can get an extra meeting:

        >>> pprint(list(iterMeetingsInTimezone(
        ...     schedule, 'UTC', date(2011, 3, 27))))
        [<Meeting on 2011-03-27 02:00 EET>,
         <Meeting on 2011-03-27 03:00 EET>,
         <Meeting on 2011-03-27 04:00 EEST>,
         <Meeting on 2011-03-27 05:00 EEST>,
         ...
         <Meeting on 2011-03-28 00:00 EEST>,
         <Meeting on 2011-03-28 01:00 EEST>,
         <Meeting on 2011-03-28 02:00 EEST>]

        >>> len(list(iterMeetingsInTimezone(
        ...     schedule, 'UTC', date(2011, 3, 27))))
        25

    Let's observe from US/Eastern, during it's dst shift:

        >>> pprint(list(iterMeetingsInTimezone(
        ...     schedule, 'US/Eastern', date(2011, 3, 12))))
        [<Meeting on 2011-03-12 07:00 EET>,
         ...
         <Meeting on 2011-03-13 06:00 EET>]

        >>> pprint(list(iterMeetingsInTimezone(
        ...     schedule, 'US/Eastern', date(2011, 3, 13))))
        [<Meeting on 2011-03-13 07:00 EET>,
         ...
         <Meeting on 2011-03-14 05:00 EET>]

        >>> pprint(list(iterMeetingsInTimezone(
        ...     schedule, 'US/Eastern', date(2011, 3, 14))))
        [<Meeting on 2011-03-14 06:00 EET>,
         ...
         <Meeting on 2011-03-15 05:00 EET>]

    """


def setUp(test=None):
    pass


def tearDown(test=None):
    pass


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF |
                   doctest.NORMALIZE_WHITESPACE)
    return unittest.TestSuite([
            doctest.DocTestSuite(optionflags=optionflags,
                                 setUp=setUp, tearDown=tearDown),
           ])



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 39.

Update person photos from File to schooltool.skin.flourish.fields.ImageFile.
"""

from zope.app.generations.utility import getRootFolder, findObjectsProviding
from zope.component.hooks import getSite, setSite

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common.fields import ImageFile


def evolvePerson(person):
    if person.photo is None:
        return
    photo = person.photo
    params = dict(photo.parameters) if photo.parameters else None
    new_photo = ImageFile(mimeType=photo.mimeType, parameters=params)
    person.photo = new_photo

    fin = photo.open("r")
    fout = new_photo.open("w")
    data = fin.read()
    fout.write(data)
    fout.close()
    fin.close()


def evolve(context):
    root = getRootFolder(context)

    old_site = getSite()
    apps = findObjectsProviding(root, ISchoolToolApplication)
    for app in apps:
        setSite(app)
        for person in app['persons'].values():
            evolvePerson(person)

    setSite(old_site)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Batching for schooltool.
"""
from zope.interface import implements
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile

from schooltool.table.interfaces import IBatch


class Batch(object):
    """Batching mechanism for Tables"""
    implements(IBatch)

    template = ViewPageTemplateFile("templates/batch.pt")

    def __init__(self, formatter, batch_size=25):
        self.formatter = formatter

        if formatter.prefix:
            self.name = "." + formatter.prefix
        else:
            self.name = ""

        self.request = self.formatter.request
        self.context = self.formatter

        item_list = list(self.context._items)
        self.full_size = len(item_list)
        self.extra_url = self.context.extra_url()
        self.base_url = self.request.URL

        self.start = int(self.request.get('batch_start' + self.name, 0))
        self.size = int(self.request.get('batch_size' + self.name, batch_size))
        if self.start >= self.full_size:
            self.start = max(0, self.full_size-self.size)
        self.length = len(item_list[self.start:self.start + self.size])

    def render(self):
        if self.size < self.full_size or self.needsBatch:
            return self.template()
        else:
            return ''

    @property
    def needsBatch(self):
        there_are_more = self.full_size > self.size + self.start
        there_were_before = self.start > 0
        return there_are_more or there_were_before

    def __len__(self):
        return self.length

    def previous_url(self):
        start = self.start - self.size
        if start >= 0:
            return '%s?batch_start%s=%s&batch_size%s=%s%s' % (
                self.base_url, self.name, start, self.name, self.size, self.extra_url)
        return None

    def next_url(self):
        start = self.size + self.start
        if self.full_size > start:
            return '%s?batch_start%s=%s&batch_size%s=%s%s' % (
                self.base_url, self.name, start, self.name, self.size, self.extra_url)

    def num(self):
        return self.start / self.size + 1

    def numBatches(self):
        num = self.full_size / self.size
        if self.full_size % self.size:
            num += 1
        return num

    def batch_urls(self):
        urls = []
        start = 0
        num = 1
        while self.full_size > start:
            css_class = 'batch_page'
            if (self.start == start):
                css_class = 'batch_page current'
            href = '%s?batch_start%s=%s&batch_size%s=%s%s' % (
                self.base_url, self.name, start, self.name, self.size, self.extra_url)
            urls.append({'href': href,
                         'num': num,
                         'class': css_class})
            num += 1
            start += self.size

        return urls


class IterableBatch(Batch):

    def __init__(self, items, request, extra_url=None, batch_size=25,
                 sort_by=None, name=None):
        self.context = None
        self.items = items
        self.request = request

        if name:
            self.name = "." + name
        else:
            self.name = ""

        def key(obj):
            try:
                return obj.get(sort_by)
            except AttributeError:
                return getattr(obj, sort_by)

        if sort_by is not None:
            item_list = sorted(items, key=key)
        else:
            item_list = items
        self.full_size = len(item_list)
        self.extra_url = extra_url or ""
        self.base_url = self.request.URL


        self.start = int(self.request.get('batch_start' + self.name, 0))
        self.size = int(self.request.get('batch_size' + self.name, batch_size))
        if self.start >= self.full_size:
            self.start = max(0, self.full_size-self.size)
        self.list = item_list[self.start:self.start + self.size]
        self.length = len(self.list)

    def __iter__(self):
        return iter(self.list)


class TokenBatch(object):
    """Another batching mechanism for Tables"""

    def __init__(self, items, start=0, size=0):
        self.items = list(items)
        self.full_size = len(self.items)
        if start >= self.full_size:
            start = max(0, self.full_size-size)
        self.start = start
        self.size = size
        self.length = len(self.items[self.start:self.start + self.size])

    @property
    def needsBatch(self):
        there_are_more = self.full_size > self.size + self.start
        there_were_before = self.start > 0
        return there_are_more or there_were_before

    def __len__(self):
        return self.length

    def __iter__(self):
        return iter(self.items[self.start:self.start + self.size])

    def previous(self):
        start = self.start - self.size
        if start >= 0:
            return {'start': start,
                    'size': self.size,
                    'items': self.items[start:start+self.size]}
        if start > -self.size:
            return {'start': 0,
                    'size': self.size,
                    'items': self.items[0:self.size]}
        return None

    def next(self):
        start = self.start + self.size
        if self.full_size > start:
            return {'start': start,
                    'size': self.size,
                    'items': self.items[start:start+self.size]}
        return None

    def current(self):
        return self.start / self.size + 1

    def numBatches(self):
        num = self.full_size / self.size
        if self.full_size % self.size:
            num += 1
        return num

    def tokens(self):
        tokens = []
        start = 0
        num = 1
        while self.full_size > start:
            tokens.append({
                    'num': num,
                    'start': start,
                    'size': self.size,
                    'current': bool(self.start == start),
                    'items': self.items[start:start+self.size]})
            num += 1
            start += self.size
        return tokens



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
csv importing.
"""
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL

import schooltool.skin.flourish.page
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.csvimport import BaseCSVImporter
from schooltool.app.browser.csvimport import BaseCSVImportView
from schooltool.app.browser.csvimport import FlourishBaseCSVImportView
from schooltool.group.interfaces import IGroupContainer
from schooltool.group.group import Group
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.skin import flourish

from schooltool.common import SchoolToolMessage as _


class GroupCSVImporter(BaseCSVImporter):
    """Group CSV Importer"""

    factory = Group

    def createAndAdd(self, data, dry_run=True):
        """Create objects and add them to the container."""

        if len(data) < 1:
            self.errors.fields.append(_('Insufficient data provided.'))
            return

        if not data[0]:
            self.errors.fields.append(_('Titles may not be empty'))
            return

        if len(data) > 1:
            description = data[1]
        else:
            description = ''

        obj = self.factory(title=data[0], description=description)
        name = self.chooser.chooseName('', obj)
        if not dry_run:
            self.container[name] = obj


class GroupCSVImportView(BaseCSVImportView):
    """View for Group CSV importer."""

    importer_class = GroupCSVImporter


class FlourishGroupCSVImportView(FlourishBaseCSVImportView, ActiveSchoolYearContentMixin):

    importer_class = GroupCSVImporter

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    def nextURL(self):
        app = ISchoolToolApplication(None)
        return self.url_with_schoolyear_id(app, view_name='groups')


class ImportGroupsLinkViewlet(flourish.page.LinkViewlet,
                              ActiveSchoolYearContentMixin):

    @property
    def url(self):
        link = self.link
        if not link:
            return None
        courses = IGroupContainer(self.schoolyear)
        return "%s/%s" % (absoluteURL(courses, self.request),
                          self.link)


class GroupMemberCSVImporter(BaseCSVImporter):
    """Group Member CSV Importer"""

    def createAndAdd(self, data, dry_run=True):
        """Create objects and add them to the container."""

        if len(data) < 1:
            self.errors.fields.append(_('Insufficient data provided.'))
            return

        if not data[0]:
            self.errors.fields.append(_('User names must not be empty.'))
            return

        app = ISchoolToolApplication(None)
        person_container = app['persons']
        username = data[0]

        if username not in person_container:
            self.errors.fields.append(_('"${username}" is not a valid username.',
                                        mapping={'username': username}))
            return

        user = person_container[username]
        if not dry_run:
            removeSecurityProxy(self.container.members).add(removeSecurityProxy(user))


class GroupMemberCSVImportView(BaseCSVImportView):
    """View for Group Member CSV importer."""

    importer_class = GroupMemberCSVImporter


class FlourishGroupMemberCSVImportView(FlourishBaseCSVImportView):

    importer_class = GroupMemberCSVImporter

    def nextURL(self):
        return absoluteURL(self.context, self.request)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for email functionality.

"""
import unittest
import doctest

from zope.interface import implements
from zope.interface.verify import verifyObject
from zope.component import provideAdapter
from zope.container.contained import NameChooser
from zope.app.testing import setup

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.email.interfaces import IEmailContainer


def doctest_EmailContainer(self):
    """Tests for toplevel container for Emails.

        >>> from zope.interface.verify import verifyObject
        >>> from schooltool.email.mail import EmailContainer
        >>> container = EmailContainer()
        >>> verifyObject(IEmailContainer, container)
        True

    It should only be able to contain emails:

        >>> from zope.container.constraints import checkObject
        >>> from schooltool.email.mail import Email
        >>> from schooltool.course.section import Section
        >>> myemail = Email('sender@example.com', ['user@example.com',],
        ...                 'Hello')
        >>> checkObject(container, 'email', myemail)
        >>> checkObject(container, 'section', Section())
        Traceback (most recent call last):
          ...
        InvalidItemType: ...

    Its attributes:

        >>> container.hostname
        >>> container.port
        >>> container.username
        >>> container.password
        >>> container.tls

    """


def doctest_Email(self):
    """Tests for Emails.

        >>> from zope.interface.verify import verifyObject
        >>> from schooltool.email.interfaces import IEmail
        >>> from schooltool.email.interfaces import IEmailContained
        >>> from schooltool.email.mail import Email
        >>> myemail = Email('sender@example.com', ['user@example.com',],
        ...                 'Hello', 'Hi, how are you?')
        >>> verifyObject(IEmail, myemail)
        True
        >>> verifyObject(IEmailContained, myemail)
        True

    Emails should only be contained in EmailContainers:

        >>> from zope.container.constraints import checkObject
        >>> from schooltool.email.mail import EmailContainer
        >>> real_container = EmailContainer()
        >>> fake_container = {}
        >>> checkObject(real_container, 'email', myemail)
        >>> checkObject(fake_container, 'email', myemail)
        Traceback (most recent call last):
          ...
        InvalidContainerType: ...

    Its attributes:

        >>> myemail.from_address
        'sender@example.com'

        >>> myemail.to_addresses
        ['user@example.com']

        >>> myemail.body
        'Hello'

        >>> myemail.subject
        'Hi, how are you?'

        >>> myemail.status_code

        >>> myemail.status_parameters
        {}

        >>> type(myemail.status_parameters)
        <class 'persistent.mapping.PersistentMapping'>

        >>> myemail.time_created
        datetime.datetime(..., tzinfo=<UTC>)

        >>> myemail.time_sent

    """


def doctest_EmailUtility():
    """Tests for EmailUtility.

        >>> class EmailContainerStub(dict):
        ...     implements(IEmailContainer)
        ...     enabled = False
        >>> email_container = EmailContainerStub()

        >>> provideAdapter(lambda app: app.emails,
        ...                adapts=(ISchoolToolApplication, ),
        ...                provides=IEmailContainer)

        >>> from schooltool.app.interfaces import IApplicationPreferences

        >>> class AppStub(object):
        ...     implements(ISchoolToolApplication, IApplicationPreferences)
        ...     def __init__(self):
        ...         self.emails = EmailContainerStub()
        ...         self.timezone = 'UTC'
        >>> app = AppStub()
        >>> provideAdapter(lambda ignored: app,
        ...                adapts=(None, ), provides=ISchoolToolApplication)


    EmailUtility is basically a wrapper around python's smtp library.

        >>> from schooltool.email.interfaces import IEmailUtility
        >>> from schooltool.email.mail import EmailUtility

        >>> util = EmailUtility()
        >>> verifyObject(IEmailUtility, util)
        True

    It's enabled when smtp_factory is specified and email container is
    enabled.

        >>> print util.smtp_factory
        None

        >>> util.smtp_factory = 'python SMTP lib'
        >>> app.emails.enabled = False

        >>> util.enabled()
        False

        >>> app.emails.enabled = True
        >>> util.enabled()
        True

        >>> util.smtp_factory = None
        >>> util.enabled()
        False

    When the utility is disabled, emails are put to the email container instead
    of being sent.

        >>> from schooltool.email.mail import Email

        >>> mail = Email('from@test',
        ...              ['to1@test', 'to2@test'], 'World', subject='Hello')

        >>> util.send(mail)
        False

        >>> app.emails
        {u'Email': <schooltool.email.mail.Email object at ...>}

        >>> print util.emailAsString(app.emails['Email'])
        Content-Type: text/plain; charset="utf-8"
        MIME-Version: 1.0
        Content-Transfer-Encoding: 7bit
        From: from@test
        To: to1@test, to2@test
        Subject: Hello
        Date: ...
        <BLANKLINE>
        World

    Putting into the email container is done via util.queue.

        >>> util.queue(mail, status_code=15, status_parameters={'foo': 1})
        >>> sorted(app.emails)
        [u'Email', u'Email-2']

    Bad mail get's status code assigned, sometimes with paramteres.

        >>> bad_mail = app.emails[u'Email-2']
        >>> bad_mail.status_code
        15

        >>> bad_mail.status_parameters
        {'foo': 1}

    Care is taken so that passed status parameters are persistent.

        >>> type(bad_mail.status_parameters)
        <class 'persistent.mapping.PersistentMapping'>

    And here is the real-life implementation of the email utility.

        >>> from schooltool.email.mail import SMTPEmailUtility
        >>> util = SMTPEmailUtility()

        >>> verifyObject(IEmailUtility, util)
        True

        >>> util.smtp_factory
        <class smtplib.SMTP at ...>

        >>> app.emails.enabled = True

        >>> util.enabled()
        True

    """


def setUp(test=None):
    setup.placefulSetUp()
    provideAdapter(NameChooser, adapts=(IEmailContainer, ))


def tearDown(test=None):
    setup.placefulTearDown()


def test_suite():
    optionflags = (doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    suite = doctest.DocTestSuite(optionflags=optionflags,
                                 setUp=setUp, tearDown=tearDown)
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
r"""
Browser views for schooltool.calendar.

iCalendar views
---------------

CalendarICalendarView can export calendars in iCalendar format

    >>> from datetime import datetime, timedelta
    >>> from schooltool.calendar.simple import ImmutableCalendar
    >>> from schooltool.calendar.simple import SimpleCalendarEvent
    >>> event = SimpleCalendarEvent(datetime(2004, 12, 16, 11, 46, 16),
    ...                             timedelta(hours=1), "doctests",
    ...                             location=u"Matar\u00f3",
    ...                             unique_id="12345678-5432@example.com")
    >>> calendar = ImmutableCalendar([event])

    >>> from zope.publisher.browser import TestRequest
    >>> view = CalendarICalendarView()
    >>> view.context = calendar
    >>> view.request = TestRequest()
    >>> output = view.show()

    >>> lines = output.splitlines(True)
    >>> from pprint import pprint
    >>> pprint(lines)
    ['BEGIN:VCALENDAR\r\n',
     'VERSION:2.0\r\n',
     'PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN\r\n',
     'BEGIN:VEVENT\r\n',
     'UID:12345678-5432@example.com\r\n',
     'SUMMARY:doctests\r\n',
     'LOCATION:Matar\xc3\xb3\r\n',
     'DTSTART:20041216T114616Z\r\n',
     'DURATION:PT1H\r\n',
     'DTSTAMP:...\r\n',
     'END:VEVENT\r\n',
     'END:VCALENDAR\r\n']

     The last line must end in '\r\n' see RFC 2445 4.4.

We can also render just the Free/Busy component of iCal

    >>> from datetime import datetime, timedelta
    >>> from schooltool.calendar.simple import ImmutableCalendar
    >>> from schooltool.calendar.simple import SimpleCalendarEvent
    >>> event = SimpleCalendarEvent(datetime(2004, 12, 16, 11, 46, 16),
    ...                             timedelta(hours=1), "doctests",
    ...                             location=u"Matar\u00f3",
    ...                             unique_id="12345678-5432@example.com")
    >>> calendar = ImmutableCalendar([event])

    >>> from zope.publisher.browser import TestRequest
    >>> view = CalendarVfbView()
    >>> view.context = calendar
    >>> view.request = TestRequest()
    >>> output = view.show()

    >>> lines = output.splitlines(True)
    >>> from pprint import pprint
    >>> pprint(lines)
    ['BEGIN:VCALENDAR\r\n',
     'VERSION:2.0\r\n',
     'PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN\r\n',
     'METHOD:PUBLISH\r\n',
     'BEGIN:VFREEBUSY\r\n',
     'FREEBUSY:20041216T114616/20041216T12461600Z\r\n',
     'END:VFREEBUSY\r\n',
     'END:VCALENDAR\r\n']

Register the iCalendar read view in ZCML as

    <browser:page
        for="schooltool.calendar.interfaces.ICalendar"
        name="calendar.ics"
        permission="zope.Public"
        class="schooltool.calendar.browser.CalendarICalendarView"
        attribute="show"
        />

Register the VfbView as calendar.vfb

    <browser:page
        for="schooltool.calendar.interfaces.ICalendar"
        name="calendar.vfb"
        permission="zope.Public"
        class="schooltool.calendar.browser.CalendarVfbView"
        attribute="show"
        />

"""

from schooltool.calendar.icalendar import convert_calendar_to_ical
from schooltool.calendar.icalendar import convert_calendar_to_vfb


class CalendarICalendarView(object):
    """RFC 2445 (ICalendar) view for calendars."""

    def show(self):
        data = "\r\n".join(convert_calendar_to_ical(self.context)) + "\r\n"
        request = self.request
        # XXX why check ?
        if request is not None:
            request.response.setHeader('Content-Type',
                                       'text/calendar; charset=UTF-8')
            request.response.setHeader('Content-Length', len(data))

        return data


class CalendarVfbView(object):
    """RFC 2445 (ICalendar) Free/Busy view for calendars."""

    def show(self):
        data = "\r\n".join(convert_calendar_to_vfb(self.context)) + "\r\n"
        request = self.request
        # XXX why check ?
        if request is not None:
            request.response.setHeader('Content-Type',
                                       'text/calendar; charset=UTF-8')
            request.response.setHeader('Content-Length', len(data))

        return data



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Selenium Functional Testing Utilities for groups.
"""

import os

from schooltool.testing.selenium import SeleniumLayer

dir = os.path.abspath(os.path.dirname(__file__))
filename = os.path.join(dir, 'stesting.zcml')

group_selenium_layer = SeleniumLayer(filename,
                                     __name__,
                                     'group_selenium_layer')

def registerSeleniumSetup():
    try:
        import selenium
    except ImportError:
        return
    from schooltool.testing import registry
    import schooltool.testing.selenium

    def addGroup(browser, schoolyear, title, **kw):
        optional = (
            'description',
            )
        browser.query.link('School').click()
        browser.query.link('Groups').click()
        browser.query.link(schoolyear).click()
        browser.query.link('Group').click()
        browser.query.name('form.widgets.title').type(title)
        for name in optional:
            if name in kw:
                value = kw[name]
                widget_id = ''.join(['form-widgets-', name])
                browser.query.id(widget_id).ui.set_value(value)
        page = browser.query.tag('html')
        browser.query.button('Submit').click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'group.add', addGroup))

    def addMembers(browser, schoolyear, group, members):
        browser.ui.group.go(schoolyear, group)
        selector = '//a[@title="Edit members for this group"]'
        browser.query.xpath(selector).click()
        selector = 'available_table-ajax-available_table--title'
        browser.query.id(selector).type(', '.join(members))
        selector = '#available_table-ajax-available_table- table'
        table = browser.query.css(selector)
        browser.query.name('SEARCH_BUTTON').click()
        browser.wait(lambda: table.expired)
        # XXX: Click Show All here in case there are lots of people
        table = browser.query.tag('table')
        browser.query.name('ADD_DISPLAYED_RESULTS').click()
        browser.wait(lambda: table.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'group.members.add', addMembers))

    def removeMembers(browser, schoolyear, group, members):
        browser.ui.group.go(schoolyear, group)
        selector = '//a[@title="Edit members for this group"]'
        browser.query.xpath(selector).click()
        # XXX: Click Show All here in case there are lots of people
        for member in members:
            selector = '#current_table-ajax-current_table- table'
            table = browser.query.css(selector)
            selector = '//button[@name="remove_item.%s"]' % member
            browser.query.xpath(selector).click()
            browser.wait(lambda: table.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'group.members.remove', removeMembers))

    def visitGroup(browser, schoolyear, group):
        browser.open('http://localhost/groups')
        browser.query.link(schoolyear).click()
        browser.query.id('SEARCH').type(group)
        table = browser.query.css('form table')
        browser.query.name('SEARCH_BUTTON').click()
        browser.wait(lambda: table.expired)
        # XXX: Click Show All here in case there are lots of groups
        sel = '//a[text()="%s"]' % group
        page = browser.query.tag('html')
        browser.query.xpath(sel).click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'group.go', visitGroup))

registerSeleniumSetup()
del registerSeleniumSetup



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Resource-related interfaces
"""

import zope.schema
from zope.container.interfaces import IContainer, IContained
from zope.container.constraints import contains, containers
from zope.interface import Interface, Attribute

from schooltool.basicperson.interfaces import IDemographics, IDemographicsFields
from schooltool.calendar.interfaces import ICalendarEvent
from schooltool.calendar.interfaces import ICalendar
from schooltool.common import SchoolToolMessage as _


class ResourceSubType(zope.schema.TextLine):
    """Sub-type of the resource is a free-form text line."""


class IBaseResource(Interface):
    """Resource."""

    title = zope.schema.TextLine(
        title=_("Title"))

    type = ResourceSubType(
        title=_("Resource Type"),
        description=_("Type of resource"),
        required=True)


    description = zope.schema.Text(
        title=_("Description"),
        required=False)

    notes = zope.schema.Text(
        title=_("Notes"),
        required=False,
        description=_("Notes for the resource."))


class IResourceContainer(IContainer):
    """Container of resources."""

    contains(IBaseResource)


class IBaseResourceContained(IBaseResource, IContained):
    """Resource contained in an IResourceContainer."""

    containers(IResourceContainer)


class IResource(IBaseResource):
    """Marker for a regular old resource"""


class ILocation(IBaseResource):
    """Location."""

    type = ResourceSubType(
        title=_("Location Type"),
        description=_("Type of location (i.e. computer lab, class room, etc.)"),
        required=True)


    capacity = zope.schema.Int(
        title=_("Capacity"),
        description=_("Capacity of the room"),
        required=False)


class IEquipment(IBaseResource):
    """Equipment."""

    type = ResourceSubType(
        title=_("Equipment Type"),
        description=_("Type of equipment (i.e. camcorder, computer, etc.)"),
        required=True)

    manufacturer = zope.schema.TextLine(
        title=_("Manufacturer"),
        description=_("Manufacturer of Equipment"),
        required=False)

    model = zope.schema.TextLine(
        title=_("Model"),
        description=_("Model of Equipment"),
        required=False)

    serialNumber = zope.schema.TextLine(
        title=_("Serial Number"),
        description=_("Serial Number of Equipment"),
        required=False)

    purchaseDate = zope.schema.Date(
        title=_("Purchase Date"),
        description=_("Purchase Date of Equipment"),
        required=False)


class IResourceTypeInformation(Interface):

    id = zope.schema.TextLine(
        title=_("Id of the resource type."),
        description=_(
            "Used for lookup of named utilities that have resource "
            "type specific information."))

    title = zope.schema.TextLine(
        title=_("The title of a resource type."),
        description=_(
            "A string that will be displayed to users in tables or "
            "resource type selection widgets."))


class IResourceFactoryUtility(Interface):

    title = zope.schema.TextLine(
        title=_("The title of a resource type."),
        description=_(
            "A string that will be displayed to users in tables or "
            "resource type selection widgets."))

    def columns():
        """Default columns for display of this resource in a table."""


class IResourceTypeSource(zope.schema.interfaces.IIterableSource):
    """Marker interface for a source of resource types."""


class IResourceSubTypes(Interface):
    """Contains a list of sub types for a given type."""

    def types():
        """returns Types among a type (subtypes)"""


class IBookResources(Interface):
    """An object that can have booked resources."""

    resources = zope.schema.Iterable(
        title=u"Booked Resources")


class IBookingCalendar(ICalendar):

    title = Attribute("")


class IBookingCalendarEvent(ICalendarEvent):
    """Event that represents a possible booking on an existing event."""


class IBookingTimetableEvent(ICalendarEvent):
    """Event that represents a possible booking in a timetable slot."""


class IResourceDemographics(IDemographics):
    """Demographics data storage for a resource

    Stores any kind of data defined by field descriptions that are set
    up for the resource container.
    """


class IResourceDemographicsFields(IDemographicsFields):
    """Demographics field storage for resource demos."""




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve28
"""
import unittest
import doctest
import os.path
import transaction

from zope.app.publication.zopepublication import ZopePublication
from zope.app.testing import setup
from ZODB.FileStorage.FileStorage import FileStorage
from ZODB.DemoStorage import DemoStorage
from ZODB.DB import DB
from zope.app.appsetup import config

import schooltool.app


class ContextStub(object):

    def __init__(self, db):
        connection = db.open()
        self.root_folder = connection.root()
        self.connection = connection


def FileStorage28():
    here = os.path.dirname(__file__)
    data_fs = os.path.join(here, "Data28.fs")
    return FileStorage(data_fs)


def doctest_evolve28():
    """Test that a database at generation 28 can be opened.

       >>> storage = DemoStorage(base=FileStorage28())
       >>> db = DB(storage, database_name="")
       >>> context = ContextStub(db)
       >>> app = context.connection.root()[ZopePublication.root_name]
       >>> from zope.component.hooks import setSite
       >>> setSite(app)

       >>> sy = app['schooltool.schoolyear'][u'2008-2009']
       >>> sy.title, sy.first, sy.last
       ('2008-2009', datetime.date(2008, 10, 1), datetime.date(2009, 1, 16))

       >>> transaction.abort()
       >>> context.connection.close()
       >>> db.close()

    """


def setUp(test):
    # Make sure unit tests are cleaned up
    setup.placefulSetUp()
    setup.placefulTearDown()
    app_module = os.path.dirname(schooltool.app.__file__)
    c = config(os.path.join(app_module, "ftesting.zcml"))


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=doctest.ELLIPSIS
                                |doctest.REPORT_ONLY_FIRST_FAILURE)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Report browser views.

"""
import datetime
import zipfile
from urllib import quote, urlencode, unquote_plus
from StringIO import StringIO

import z3c.form
import zope.contentprovider.interfaces
import zope.event
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import adapts, queryMultiAdapter
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.interface import Interface
from zope.publisher.browser import BrowserView
from zope.publisher.browser import BrowserPage
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.interfaces import NotFound
from zope.publisher.interfaces import IPublishTraverse
from zope.traversing.browser.absoluteurl import AbsoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.interface import implements
from zope.cachedescriptors.property import Lazy
from zope.proxy import getProxiedObject
from z3c.form import button

import schooltool.traverser.traverser
from schooltool.common import format_message
from schooltool.person.interfaces import IPerson
from schooltool.report.interfaces import IReportLinksURL
from schooltool.report.interfaces import IReportLinkViewlet
from schooltool.report.interfaces import IReportFile
from schooltool.report.interfaces import IArchivePage
from schooltool.report.report import IFlourishReportLinkViewletManager
from schooltool.report.report import getReportRegistrationUtility
from schooltool.report.report import ReportTask
from schooltool.report.report import ArchiveReportTask
from schooltool.skin import flourish
from schooltool.skin.flourish.page import WideContainerPage
from schooltool.skin.flourish.page import RefineLinksViewlet
from schooltool.skin.flourish import IFlourishLayer
from schooltool.skin.flourish.form import DialogForm
from schooltool.task.tasks import query_message
from schooltool.task.interfaces import IRemoteTask
from schooltool.task.progress import TaskProgress
from schooltool.task.progress import normalized_progress
from schooltool.task.browser.task import MessageDialog

from schooltool.common import SchoolToolMessage as _


class ReportsView(BrowserView):

    template = ViewPageTemplateFile('templates/report_links.pt')

    def __call__(self):
        return self.template()


class StudentReportsView(ReportsView):

    title = _('Student Reports')


class GroupReportsView(ReportsView):

    title = _('Group Reports')


class SchoolYearReportsView(ReportsView):

    title = _('School Year Reports')


class TermReportsView(ReportsView):

    title = _('Term Reports')


class SectionReportsView(ReportsView):

    title = _('Section Reports')


class FlourishReportsView(flourish.page.Page):
    """Report request view base class."""
    content_template = ViewPageTemplateFile('templates/f_report_links.pt')
    subtitle = _('Reports')


def reportLinksURL(ob, request, name=''):
    """Helper method to obtain the report links URL"""
    url = queryMultiAdapter((ob, request), IReportLinksURL, name=name)
    if url is None:
       return ''
    return url


class ReportReferenceView(BrowserView):

    template = ViewPageTemplateFile('templates/report_reference.pt')

    def __call__(self):
        return self.template()

    def rows(self):
        collator = ICollator(self.request.locale)
        utility = getReportRegistrationUtility()
        app = self.context

        rows = []
        for group_key, group_reports in utility.reports_by_group.items():
            reference_url = reportLinksURL(app, self.request, name=group_key)
            for report in group_reports:
                row = {
                    'url': reference_url,
                    'group': report['group'],
                    'title': report['title'],
                    'description': report['description'],
                    }
                rows.append([collator.key(report['group']),
                             collator.key(report['title']),
                             row])

        return [row for group, title, row in sorted(rows)]


class FlourishReportReferenceView(WideContainerPage, ReportReferenceView):

    def done_link(self):
        return absoluteURL(self.context, self.request) + '/manage'


    def rows(self):
        self.collator = ICollator(self.request.locale)
        utility = getReportRegistrationUtility()
        app = self.context
        rows = {}
        for group_key, group_reports in utility.reports_by_group.items():
            reference_url = reportLinksURL(app, self.request, name=group_key)
            for report in group_reports:
                group = report['group']
                name = report['name']
                row = {
                    'url': reference_url,
                    'group': group,
                    'title': report['title'],
                    'file_type': report['file_type'].upper(),
                    'description': report['description'],
                    }
                # XXX: this check is needed to override old skin
                #      report links with flourish ones
                if (group, name) not in rows or \
                   report['layer'] is IFlourishLayer:
                    rows[group, name] = row
        return sorted(rows.values(), key=self.sortKey)

    def sortKey(self, row):
        return self.collator.key(row['group']), self.collator.key(row['title'])


class ReportsLinks(RefineLinksViewlet):
    """Reports links viewlet."""

    implements(IFlourishReportLinkViewletManager)

    body_template = ViewPageTemplateFile('templates/f_report_links_body.pt')

    @Lazy
    def items(self):
        items = super(ReportsLinks, self).items
        result = []
        for item in items:
            viewlet = item['viewlet']
            url = viewlet.link
            is_report_link = bool(getattr(viewlet, 'file_type', ''))
            if is_report_link:
                file_type = translate(viewlet.file_type,
                                      context=self.request)
                description = translate(viewlet.description,
                                        context=self.request)
                querystring = urlencode({
                        'file_type': file_type.encode('utf-8').upper(),
                        'description': description.encode('utf-8')})
                url = '%s?%s' % (viewlet.report_link, querystring)
            result.append({
                    'class': item['class'],
                    'viewlet': viewlet,
                    'content': item['content'],
                    'is_report_link': is_report_link,
                    'link_id': viewlet.link.replace('.', '_'),
                    'form_id': viewlet.link.replace('.', '_') + '_form',
                    'title': translate(viewlet.title, context=self.request),
                    'url': url,
                    })
        return result


class ReportLinkViewletProxy(flourish.viewlet.ViewletProxy):

    adapts(IReportLinkViewlet)

    def __init__(self, *args, **kw):
        super(ReportLinkViewletProxy, self).__init__(*args, **kw)
        self.restoreSortingAttrs()

    def restoreSortingAttrs(self):
        unproxied = getProxiedObject(self)
        self.before = getattr(unproxied, 'before', ())
        self.after = getattr(unproxied, 'after', ())


class RequestReportDownloadDialog(DialogForm, z3c.form.form.EditForm):

    template = ViewPageTemplateFile('templates/f_request_report_download.pt')

    dialog_submit_actions = ('download',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_("Download"), name='download')
    def handleDownload(self, action):
        self.request.response.redirect(self.nextURL())
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(RequestReportDownloadDialog, self).updateActions()
        self.actions['download'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        raise NotImplementedError("nextURL must redirect to a 'downloadable' view")

    @property
    def file_type(self):
        if 'file_type' in self.request:
            return unquote_plus(self.request['file_type'])

    @property
    def description(self):
        if 'description' in self.request:
            return unquote_plus(self.request['description'])


class RequestRemoteReportDialog(RequestReportDownloadDialog):

    template = ViewPageTemplateFile('templates/f_request_report_download.pt')

    dialog_submit_actions = ('download',)
    dialog_close_actions = ('cancel',)
    label = None

    report_builder = None # report generating class or view name

    task_factory = ReportTask

    report_task = None
    replace_dialog = None

    fields = z3c.form.field.Fields(Interface)
    form_params = None

    def resetForm(self):
        self.form_params = {}

    def update(self):
        self.resetForm()
        RequestReportDownloadDialog.update(self)

    def getContent(self):
        return self.form_params

    @button.buttonAndHandler(_("Generate"), name='download')
    def handleDownload(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changes = self.applyChanges(data)
        if changes:
            self.status = self.successMessage
        else:
            self.status = self.noChangesMessage
        task = self.task_factory(self.report_builder, self.target)
        self.schedule(task)
        self.report_task = task
        message = query_message(task)
        if message is None:
            self.ajax_settings['dialog'] = 'close'
            return
        content = queryMultiAdapter(
            (message, self.request), name='dialog')
        if content is None:
            self.ajax_settings['dialog'] = 'close'
            return
        self.replace_dialog = content

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(RequestReportDownloadDialog, self).updateActions()
        self.actions['download'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @property
    def target(self):
        return self.context

    def updateTaskParams(self, task):
        """Subclasses should update task.request_params dict here."""
        task.request_params.update(self.form_params)

    def schedule(self, task):
        self.updateTaskParams(task)
        task.schedule(self.request)

    def render(self, *args, **kw):
        if self.replace_dialog is not None:
            return self.replace_dialog.render(*args, **kw)
        return super(RequestRemoteReportDialog, self).render(*args, **kw)


class ReportAbsoluteURLAdapter(AbsoluteURL):
    adapts(IReportFile, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        base = absoluteURL(self.context.__parent__, self.request)
        filename = quote(unicode(self.context.__name__).encode('UTF-8'))
        url = base + '/download/' + filename
        return url

    __call__ = __str__


class ReportMessageDownloads(object):
    implements(IPublishTraverse)
    file_attrs = ('report',)
    files = None

    def __init__(self, context, request, name='download'):
        self.__name__ = name
        self.__parent__ = context
        self.context = context
        self.request = request
        self.collectFiles()

    def collectFiles(self):
        self.files = {}
        for attr in self.file_attrs:
            value = getattr(self.context, attr, None)
            if value is None:
                continue
            self.files[value.__name__] = value

    def publishTraverse(self, request, name):
        if name not in self.files:
            raise NotFound(self.context, name, self.request)
        return self.files[name]


class ReportMessageDownloadsPlugin(schooltool.traverser.traverser.TraverserPlugin):

    def traverse(self, name):
        return ReportMessageDownloads(self.context, self.request)


class DownloadReportDialog(MessageDialog):

    template = flourish.templates.File('templates/f_download_report_dialog.pt')

    @property
    def report(self):
        return getattr(self.context, 'report', None)

    @property
    def report_generated(self):
        return bool(self.report)

    @property
    def main_recipient(self):
        person = IPerson(self.request, None)
        if self.context.recipients is None:
            return None
        recipients = sorted(self.context.recipients, key=lambda r: r.__name__)
        if person in recipients:
            return person
        for recipient in recipients:
            if flourish.canView(recipient):
                return recipient
        return None

    @Lazy
    def failure_ticket_id(self):
        sender = self.context.sender
        if (IRemoteTask.providedBy(sender) and
            sender.failed):
            return sender.__name__
        return None


class ShortReportMessage(flourish.content.ContentProvider):

    @Lazy
    def failure_ticket_id(self):
        sender = self.context.sender
        if (IRemoteTask.providedBy(sender) and
            sender.failed):
            return sender.__name__
        return None


class ProgressReportPage(BrowserPage):

    render_invariant = False
    render_debug = False

    base_filename = 'export'

    message_title = _('export')

    task_progress = None
    overall_line_id = None

    def __init__(self, context, request):
        super(ProgressReportPage, self).__init__(context, request)
        self.makeProgress()

    def makeFileName(self, basename):
        if self.render_invariant:
            return '%s.xls' % basename
        timestamp = datetime.datetime.now().strftime('%y-%m-%d-%H-%M')
        return '%s_%s.xls' % (basename, timestamp)

    @property
    def filename(self):
        return self.makeFileName(self.base_filename)

    def makeProgress(self):
        task_id = getattr(self.request, 'task_id', None)
        self.task_progress = TaskProgress(task_id)

    def update(self):
        self.makeProgress()

    def updateOverall(self):
        if self.overall_line_id is None:
            return
        max_progress = 0.0
        overall = 0.0
        for lid, line in self.task_progress.lines.items():
            if (line.get('progress', None) is not None and
                lid != self.overall_line_id):
                overall += line.progress
                max_progress += 1.0
        if max_progress > 0.0:
            overall /= max_progress
        self.task_progress('overall', progress=overall)

    def progress(self, importer, value):
        self.task_progress(importer, progress=value, active=True)
        self.updateOverall()

    def finish(self, importer):
        self.task_progress.finish(importer)
        self.updateOverall()


class ArchiveFileManager(flourish.viewlet.ViewletManagerBase):

    collecting_title = _('Collecting (${name})')
    overall_line_id = 'overall'

    @property
    def title(self):
        return getattr(self.view, 'message_title', _("Archive"))

    @property
    def task_progress(self):
        return self.view.task_progress

    def addArchivers(self):
        progress = self.task_progress
        progress.add('active_file', title=None, active=False)
        if self.overall_line_id is not None:
            progress.add(self.overall_line_id,
                         title=self.title, progress=0.0)
            self.view.overall_line_id = self.overall_line_id

    def update(self):
        self.task_progress.add('collecting',
            title=u'', progress=0.0)
        event = zope.contentprovider.interfaces.BeforeUpdateEvent
        total = len(self.viewlets)
        for nv, viewlet in enumerate(self.viewlets):
            title = format_message(self.collecting_title,
                                   {'name': viewlet.title or _('files')})
            self.task_progress.force(
                'collecting',
                title=title,
                progress=normalized_progress(nv, total),
                active=True)
            zope.event.notify(event(viewlet, self.request))
            viewlet.update()
        title = format_message(self.collecting_title,
                               {'name': _('files')})
        self.task_progress.force('collecting', title=title, progress=None, active=False)
        self.addArchivers()

    def render(self, archive):
        for viewlet in self.viewlets:
            viewlet(archive)
        if self.overall_line_id is not None:
            self.task_progress.force(self.overall_line_id, progress=1.0)

    def updateOverall(self):
        self.view.updateOverall()

    def progress(self, archiver_name, filename, progress):
        self.task_progress(archiver_name, progress=progress, active=True)
        self.task_progress('active_file', title=filename, active=True, progress=None)
        self.updateOverall()

    def finish(self, archiver_name):
        self.task_progress.finish(archiver_name)
        self.task_progress('active_file', title=u'', active=True)
        self.updateOverall()


class FileArchiver(flourish.viewlet.Viewlet):

    title = None

    def addArchivers(self):
        progress = self.view.task_progress
        title = self.title or _('Reports')
        progress.add(self.__name__, active=False,
                     title=title, progress=0.0)

    def setTitle(self, title):
        self.title = title
        progress = self.view.task_progress
        progress(self.__name__, title=title)

    def progress(self, filename, file_n, total_files, file_progress=0.0):
        self.manager.progress(
            self.__name__,
            filename,
            normalized_progress(
                file_n, total_files,
                file_progress, 1.0,
                ))

    def finish(self):
        self.manager.finish(self.__name__)

    def update(self):
        flourish.viewlet.Viewlet.update(self)
        self.addArchivers()

    def render(self, archive):
        self.finish()

    def queryView(self, item, request, view_name):
        renderer = zope.component.queryMultiAdapter(
            (item, request), name=view_name)
        return renderer


class ReportArchivePage(ProgressReportPage):
    implements(IArchivePage)

    compression = zipfile.ZIP_STORED

    collecting_title = _("Preparing archive")
    archiving_title = _("Archiving")
    finished_title = _("Archive complete")

    def makeFileName(self, basename):
        if self.render_invariant:
            return '%s.zip' % basename
        timestamp = datetime.datetime.now().strftime('%y-%m-%d-%H-%M')
        return '%s_%s.zip' % (basename, timestamp)

    def render(self, stream, *args, **kw):
        """Render a zip to stream, return number of files written."""
        file_manager = flourish.content.queryContentProvider(
            self.context, self.request, self, 'file_manager')
        if file_manager is None:
            return 0

        self.task_progress.title = self.collecting_title
        file_manager.update()

        self.task_progress.title = self.archiving_title
        written = 0
        zip_file = zipfile.ZipFile(stream, 'w', compression=self.compression)
        try:
            file_manager.render(zip_file)
            written = len(zip_file.filelist)
        finally:
            zip_file.close()
        self.task_progress.title = self.finished_title
        return written

    def __call__(self, *args, **kw):
        self.update()
        if self.request.response.getStatus() in [300, 301, 302, 303,
                                                 304, 305, 307]:
            return u''
        stream = StringIO()
        self.render(stream, *args, **kw)
        data = stream.getvalue()
        if not data:
            return u''
        response = self.request.response
        response.setHeader('Content-Type', 'application/zip')
        response.setHeader('Content-Length', len(data))
        return data


class RequestReportArchiveDialog(RequestRemoteReportDialog):

    task_factory = ArchiveReportTask


class ReportReferenceLinkViewlet(flourish.page.LinkViewlet):
    pass



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Interface definitions for SchoolTool calendars.

There are two interfaces for calendars: `ICalendar` for read-only calendars,
and `IEditCalendar` for read-write calendars.

Semantically calendars are unordered sets of events.
"""

import zope.schema
from zope.interface import Interface, Attribute
from zope.location.interfaces import ILocation, IContained


Unchanged = object() # marker


class ICalendar(Interface):
    """Calendar.

    A calendar is a set of calendar events (see ICalendarEvent).  Recurring
    events are listed only once.
    """

    def __len__():
        """Return the number of events in this calendar."""

    def __iter__():
        """Return an iterator over all events in this calendar.

        The order of events is not defined.
        """

    def find(unique_id):
        """Return an event with the given unique id.

        Raises a KeyError if there is no event with this id.
        """

    def expand(first, last):
        """Return an iterator over all expanded events in a given time period.

        "Expanding" here refers to expanding recurring events, that is,
        creating objects for all occurrences of recurring events.  If a
        recurring event has occurreces that overlap the specified time
        interval, every such occurrence is represented as a new calendar event
        with the `dtstart` attribute replaced with the date and time of that
        occurrence.  These events provide IExpandedCalendarEvent and have an
        additional attribute which points to the original event.  Note that
        only recurring events are replaced by proxies, nonrecurring events
        are returned as-is.

        `first` and `last` are datetime.datetimes and define a half-open
        time interval.

        The order of returned events is not defined.
        """


class IEditCalendar(ICalendar):
    """Editable calendar.

    Calendar events are read-only, so to change an event you need to remove
    the old event, and add a replacement event in the calendar.
    """

    def clear():
        """Remove all events."""

    def addEvent(event):
        """Add an event to the calendar.

        Raises ValueError if an event with the same unique_id already exists
        in the calendar.

        Returns the newly added event (which may be a copy of the argument,
        e.g. if the calendar needs its events to be instances of a particular
        class).

        It is perhaps not a good idea to add calendar events that have no
        occurrences into calendars (see ICalendarEvent.hasOccurrences), as they
        will be invisible in date-based of calendar views.

        Do not call addEvent while iterating over the calendar.
        """

    def removeEvent(event):
        """Remove event from the calendar.

        Raises ValueError if event is not present in the calendar.

        Do not call removeEvent while iterating over the calendar.
        """


class IRecurrenceRule(Interface):
    """Base interface of the recurrence rules.

    Recurrence rules are stored as attributes of ICalendarEvent.  They
    are also immutable and comparable.  To modify the recurrence
    rule of an event, you need to create a new recurrence rule, and a new
    event:

        replacement_rule = event.recurrence.replace(count=3, until=None)
        replacement_event = event.replace(recurrence=replacement_rule)
        calendar.removeEvent(event)
        calendar.addEvent(replacement_event)

    """

    interval = zope.schema.Int(
        title=u"Interval",
        min=1,
        description=u"""
        Interval of recurrence (a positive integer).

        For example, to indicate that an event occurs every second day,
        you would create a DailyRecurrenceRule witl interval equal to 2.
        """)

    count = zope.schema.Int(
        title=u"Count",
        required=False,
        description=u"""
        Number of times the event is repeated.

        Can be None or an integer value.  If count is not None then
        until must be None.  If both count and until are None the
        event repeats forever.
        """)

    until = zope.schema.Date(
        title=u"Until",
        required=False,
        description=u"""
        The date of the last recurrence of the event.

        Can be None or a datetime.date instance.  If until is not None
        then count must be None.  If both count and until are None the
        event repeats forever.
        """)

    exceptions = zope.schema.List(
        title=u"Exceptions",
        value_type=zope.schema.Date(),
        description=u"""
        A list of days when this event does not occur.

        Values in this list must be instances of datetime.date.
        """)

    def replace(interval=Unchanged, count=Unchanged, until=Unchanged,
                exceptions=Unchanged):
        """Return a copy of this recurrence rule with new specified fields."""

    def __eq__(other):
        """See if self == other."""

    def __ne__(other):
        """See if self != other."""

    def apply(event, startdate=None, enddate=None):
        """Apply this rule to an event.

        This is a generator that returns the dates on which the event should
        recur.  Be careful when iterating over these dates -- rules that do not
        have either 'until' or 'count' attributes will go on forever.

        The optional enddate attribute can be used to set a range on the dates
        generated by this function (inclusive).
        """

    def iCalRepresentation(dtstart):
        """Return the rule in iCalendar format.

        Returns a list of strings, each corresponding to a line of iCalendar.
        The first line is for the recurrence rule itself, the following ones
        (if any) describe exception dates.

        dtstart is a datetime representing the date that the recurring
        event starts on.
        """


class IDailyRecurrenceRule(IRecurrenceRule):
    """Daily recurrence."""


class IYearlyRecurrenceRule(IRecurrenceRule):
    """Yearly recurrence."""


class IWeeklyRecurrenceRule(IRecurrenceRule):
    """Weekly recurrence."""

    weekdays = zope.schema.Set(
        title=u"Weekdays",
        value_type=zope.schema.Int(min=0, max=6),
        description=u"""
        A set of weekdays when this event occurs.

        Weekdays are represented as integers from 0 (Monday) to 6 (Sunday).
        This is what the `calendar` and `datetime` modules use.

        The event repeats on the weekday of the first occurence even
        if that weekday is not in this set.
        """)

    def replace(interval=Unchanged, count=Unchanged, until=Unchanged,
                exceptions=Unchanged, weekdays=Unchanged):
        """Return a copy of this recurrence rule with new specified fields."""


class IMonthlyRecurrenceRule(IRecurrenceRule):
    """Monthly recurrence."""

    monthly = zope.schema.Choice(
        title=u"Type",
        values=('monthday', 'weekday', 'lastweekday'),
        description=u"""
        Specification of a monthly occurence.

        Can be one of three values: 'monthday', 'weekday', 'lastweekday'.

        'monthday'    specifies that the event recurs on the same day of month
                      (e.g., 25th day of a month).

        'weekday'     specifies that the event recurs on the same week
                      within a month on the same weekday, indexed from the
                      first (e.g. 3rd Friday of a month).

        'lastweekday' specifies that the event recurs on the same week
                      within a month on the same weekday, indexed from the
                      end of month (e.g. 2nd last Friday of a month).
        """)

    def replace(interval=Unchanged, count=Unchanged, until=Unchanged,
                exceptions=Unchanged, monthly=Unchanged):
        """Return a copy of this recurrence rule with new specified fields."""


class ICalendarEvent(Interface):
    """A calendar event.

    Calendar events are comparable.

    Events are compared in chronological order, so lists of events can be
    sorted.  If two events start at the same time, they are ordered according
    to their titles.

    While `unique_id` is a globally unique ID of a calendar event, you can
    have several calendar event objects with the same value of `unique_id`,
    and they will not be equal if any their attributes are different.
    Semantically these objects are different versions of the same calendar
    event.

    """

    unique_id = zope.schema.TextLine(
        title=u"UID",
        description=u"""
        A globally unique id for this calendar event.

        iCalendar (RFC 2445) recommeds using the RFC 822 addr-spec syntax
        for unique IDs.  Put the current timestamp and a random number
        on the left of the @ sign, and put the hostname on the right.
        """)

    dtstart = zope.schema.Datetime(
        title=u"Start",
        description=u"""
        Date and time when the event starts.
        """)

    duration = zope.schema.Timedelta(
        title=u"Duration",
        description=u"""
        The duration of the event (datetime.timedelta).

        You can compute the event end date/time by adding duration to dtstart.
        """)

    title = zope.schema.TextLine(
        title=u"Title",
        description=u"""The title of the event.""")

    description = zope.schema.Text(
        title=u"Description",
        required=False,
        description=u"""A description of the event.""")

    location = zope.schema.TextLine(
        title=u"Location",
        required=False,
        description=u"""The location where this event takes place.""")

    recurrence = zope.schema.Object(
        title=u"Recurrence",
        schema=IRecurrenceRule,
        required=False,
        description=u"""
        The recurrence rule, if this is a recurring event, otherwise None.
        """)

    allday = zope.schema.Bool(
        title=u"All Day Event",
        required=False)

    def __eq__(other):
        """See if self == other."""

    def __ne__(other):
        """See if self != other."""

    def __lt__(other):
        """See if self < other."""

    def __gt__(other):
        """See if self > other."""

    def __le__(other):
        """See if self <= other."""

    def __ge__(other):
        """See if self >= other."""

    def hasOccurrences():
        """Does the event have any occurrences?

        Normally all events have at least one occurrence.  However if you have
        a repeating event that repeats a finite number of times, and all those
        repetitions are listed as exceptions, then hasOccurrences() will return
        False.  There are other corner cases as well (e.g. a recurring event
        with until date that is earlier than dtstart).
        """

    def expand(first, last):
        """Return an iterator over all expanded events in a given time period.

        "Expanding" here refers to expanding recurring events, that
        is, creating objects for all occurrences of this event if it
        is a recurring event.  If a recurring event has occurreces
        that overlap the specified time interval, every such
        occurrence is represented as a new calendar event with the
        `dtstart` attribute replaced with the date and time of that
        occurrence.  These events provide IExpandedCalendarEvent and
        have an additional attribute which points to the original
        event.  Note that only recurring events are replaced by
        proxies, nonrecurring events are returned as-is.

        `first` and `last` are datetime.datetimes and define a half-open
        time interval.

        The order of returned events is not defined.
        """


class IExpandedCalendarEvent(ICalendarEvent):
    """A single occurrence of a recurring calendar event.

    The original event is stored in the `original` attribute.  The `dtstart`
    attribute contains the date and time of this occurrence and may differ
    from the `dtstart` attribute of the original event.  All other attributes
    are the same.
    """

    dtstart = zope.schema.Datetime(
        title=u"Start",
        description=u"""
        Date and time when this occurrence of the event starts.
        """)

    original = zope.schema.Object(
        title=u"Original",
        schema=ICalendarEvent,
        description=u"""
        The recurring event that generated this occurrence.
        """)


class IWriteCalendar(Interface):

    def write(data, charset='UTF-8'):
        """Update the calendar data
        """


class ISchoolToolCalendar(IEditCalendar, ILocation):
    """A SchoolTool calendar.

    Calendars stored within all provide ISchoolToolCalendarEvent.
    """

    title = zope.schema.TextLine(
        title=u"Title",
        description=u"Title of the calendar.")


class ISchoolToolCalendarEvent(ICalendarEvent, IContained):
    """An event that is contained in a SchoolTool calendar."""

    resources = Attribute(
        """Resources that are booked by this event""")

    owner = Attribute(
        """Object which this event belongs to.""")

    def bookResource(resource):
        """Book a resource."""

    def unbookResource(resource):
        """Book a resource."""


class IHaveCalendar(Interface):
    """Marker interface for components that can have calendars.

    Components providing this interface are adaptable to
    ``ISchoolToolCalendar``.
    """



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Catalog tests
"""
import unittest
import doctest

from pprint import pprint

from zope.intid.interfaces import IIntIds
from zope.component import provideUtility
from zope.component import provideAdapter

from schooltool.app.browser.testing import setUp, tearDown


def doctest_FilterImplementing():
    """Doctest for FilterImplementing

        >>> from zope.interface import implements, Interface
        >>> class IFoo(Interface):
        ...     pass
        >>> class Foo(object):
        ...     implements(IFoo)

        >>> from schooltool.table.catalog import FilterImplementing
        >>> filter = FilterImplementing(IFoo)
        >>> filter('index', 'docid', object())
        False
        >>> filter('index', 'docid', Foo())
        True

    """


def doctest_IndexedFilterWidget():
    """Doctest for IndexedFilterWidget.

    First we need a container with a catalog associated with it:

        >>> from zope.catalog.interfaces import ICatalog
        >>> from zope.interface import implements
        >>> class IndexStub(object):
        ...     def __init__(self):
        ...         self.documents_to_values = {}

    Our catalog has an index for title attribute:

        >>> class CatalogStub(dict):
        ...     implements(ICatalog)
        ...     def __init__(self):
        ...         self['title'] = IndexStub()

        >>> catalog = CatalogStub()

        >>> class ContainerStub(object):
        ...     def __conform__(self, iface):
        ...         if iface == ICatalog:
        ...             return catalog

    Let's create the IndexedFilterWidget:

        >>> from zope.publisher.browser import TestRequest
        >>> from schooltool.table.catalog import IndexedFilterWidget

        >>> container = ContainerStub()
        >>> request = TestRequest()
        >>> widget = IndexedFilterWidget(container, request)

    Indexed filter widgets do not get items themselves, as filtering
    through normal object attributes would be just too slow, instead
    they get dicts with item int ids in them. All the information used
    for filtering is stored in the catalog associated with the context
    container:

        >>> catalog['title'].documents_to_values[5] = 'Lambda'
        >>> catalog['title'].documents_to_values[6] = 'Alpha'
        >>> catalog['title'].documents_to_values[7] = 'Beta'

        >>> items = [{'id': 5}, {'id': 6}, {'id': 7}]
        >>> request.form = {'SEARCH': 'lamb'}
        >>> widget.filter(items)
        [{'id': 5}]

   The search is case insensitive:

        >>> request.form = {'SEARCH': 'AlphA'}
        >>> widget.filter(items)
        [{'id': 6}]

    If clear search button is clicked, the form attribute is cleared,
    and all items are displayed:

        >>> request.form['CLEAR_SEARCH'] = 'Yes'

        >>> widget.filter(items)
        [{'id': 5}, {'id': 6}, {'id': 7}]
        >>> request.form['SEARCH']
        ''

    """


def doctest_IndexedGetterColumn():
    """Tests for IndexedGetterColumn.

    Indexed column requires an additional keyword argument (index)
    for it's constructor:

        >>> from schooltool.table.column import IndexedGetterColumn
        >>> column = IndexedGetterColumn(index='title', getter=lambda i, f: i.title)
        >>> column.index
        'title'

    Items used by this column are not ordinary objects, but rather
    index dicts:


        >>> class IndexStub(object):
        ...     def __init__(self):
        ...         self.documents_to_values = {}
        >>> index = IndexStub()
        >>> index.documents_to_values[5] = 'Peter'

        >>> int_ids = {}
        >>> class IntIdsStub(object):
        ...     def getObject(self, id):
        ...         return int_ids[id]
        >>> provideUtility(IntIdsStub(), IIntIds)

        >>> catalog = {'title': index}
        >>> item = {'id': 5,
        ...         'catalog': catalog}

    The key of the index is used when getting the sort key of the
    column, so the only object touched is the index in the catalog,
    the Peter object is not used when sorting in any way:

        >>> column.getSortKey(item, None)
        'Peter'

    """


def doctest_IndexedLocaleAwareGetterColumn():
    """Tests for IndexedLocaleAwareGetterColumn.

    Create a formatter with a request:

        >>> from zope.publisher.browser import TestRequest
        >>> request = TestRequest()
        >>> class FormatterStub(object):
        ...     request = request
        >>> formatter = FormatterStub()

    Register collation adapter:

        >>> from zope.i18n.interfaces.locales import ICollator
        >>> from zope.i18n.interfaces.locales import ILocale
        >>> from zope.interface import implements
        >>> from zope.component import adapts
        >>> class CollatorAdapterStub(object):
        ...     implements(ICollator)
        ...     adapts(ILocale)
        ...     def __init__(self, context):
        ...         self.context = context
        ...     def key(self, string):
        ...         return "CollatorKey(%s)" % string
        >>> provideAdapter(CollatorAdapterStub)

    Indexed column requires an additional keyword argument (index)
    for it's constructor:

        >>> from schooltool.table.column import IndexedLocaleAwareGetterColumn
        >>> column = IndexedLocaleAwareGetterColumn(index='title',
        ...                                         getter=lambda i, f: i.title)
        >>> column.index
        'title'

    Items used by this column are not ordinary objects, but rather
    index dicts:

        >>> class IndexStub(object):
        ...     def __init__(self):
        ...         self.documents_to_values = {}
        >>> index = IndexStub()
        >>> index.documents_to_values[5] = 'Peter'

        >>> context = {}
        >>> catalog = {'title': index}
        >>> item = {'id': 5,
        ...         'catalog': catalog}

    The sort key for this column is not the title of the object, but
    rather a collator key derived from the title:

        >>> column.getSortKey(item, formatter)
        'CollatorKey(Peter)'

    """


def doctest_IndexedTableFormatter_columns():
    """Tests for IndexedTableFormatter.columns.

    The default column for an indexed table formatter is an indexed
    getter column set to display the title of objects:

        >>> from schooltool.table.catalog import IndexedTableFormatter
        >>> formatter = IndexedTableFormatter(None, None)
        >>> columns = formatter.columns()
        >>> columns
        [<schooltool.table.column.IndexedGetterColumn object at ...>]

        >>> columns[0].title
        u'Title'
        >>> columns[0].index
        'title'

    """


def doctest_IndexedTableFormatter_items():
    """Tests for IndexedTableFormatter.items.

        >>> from zope.catalog.interfaces import ICatalog
        >>> from zope.interface import implements
        >>> from zc.catalog.interfaces import IExtentCatalog
        >>> from schooltool.table.catalog import IndexedTableFormatter

        >>> class IndexStub(object):
        ...     def __init__(self):
        ...         self.documents_to_values = {}

        >>> class CatalogStub(dict):
        ...     implements(ICatalog)
        ...     def __init__(self, index):
        ...         self['__name__'] = index
        ...     def __repr__(self):
        ...         return '<Catalog>'

        >>> class ExtentCatalogStub(object):
        ...     implements(IExtentCatalog)
        ...     def __init__(self, extent):
        ...         self.extent = extent
        ...     def __repr__(self):
        ...         return '<ExtentCatalog>'

        >>> class ContainerStub(object):
        ...     def __init__(self, catalog):
        ...         self.catalog = catalog
        ...     def __conform__(self, iface):
        ...         if iface == ICatalog:
        ...             return self.catalog
        ...     def __repr__(self):
        ...         return '<Container>'

    The IndexedTableFormatter relies on catalog to list all the items.

    If catalog implements IExtentCatalog, it's extent is listed to obtain indexed
    item ids.

        >>> catalog = []
        >>> catalog = ExtentCatalogStub([])
        >>> formatter = IndexedTableFormatter(ContainerStub(catalog), None)
        >>> formatter.items()
        []

        >>> catalog.extent[:] = [1, 2, 3]
        >>> formatter.items()
        [{'catalog': <ExtentCatalog>,
          'id': 1},
         {'catalog': <ExtentCatalog>,
          'id': 2},
         {'catalog': <ExtentCatalog>,
          'id': 3}]

    For other catalogs, ids indexed in first index are used.

        >>> from schooltool.table.catalog import IndexedTableFormatter
        >>> index = IndexStub()
        >>> formatter = IndexedTableFormatter(
        ...     ContainerStub(CatalogStub(index)), None)
        >>> formatter.items()
        []

    Let's index some objects:

        >>> index.documents_to_values[1] = 'peter'
        >>> index.documents_to_values[2] = 'john'
        >>> index.documents_to_values[3] = 'ted'

    Now we should get a list of index dicts:

        >>> pprint(formatter.items())
        [{'catalog': <Catalog>,
          'id': 1},
         {'catalog': <Catalog>,
          'id': 2},
         {'catalog': <Catalog>,
          'id': 3}]

    """


def doctest_IndexedTableFormatter_indexItems():
    """Tests for IndexedTableFormatter.indexItems.

        >>> from zope.catalog.interfaces import ICatalog
        >>> class ContainerStub(object):
        ...     def __conform__(self, iface):
        ...         if iface == ICatalog:
        ...             return '<Catalog>'
        ...     def __repr__(self):
        ...         return '<Container>'
        >>> from schooltool.table.catalog import IndexedTableFormatter
        >>> formatter = IndexedTableFormatter(ContainerStub(), None)

    The formatter has a function that converts a list of objects into
    a list of index dicts that can be displayed by Indexed columns and
    filtered by indexed filter widgets:

        >>> class IntIdsStub(object):
        ...     def getId(self, obj):
        ...         return obj.id
        >>> provideUtility(IntIdsStub(), IIntIds)

        >>> class ItemStub(object):
        ...     def __init__(self, name, id):
        ...         self.__name__, self.id = name, id

        >>> items = [ItemStub('pete', 1), ItemStub('john', 2)]
        >>> pprint(formatter.indexItems(items))
        [{'catalog': '<Catalog>',
          'id': 1},
         {'catalog': '<Catalog>',
          'id': 2}]

    """


def doctest_IndexedTableFormatter_setUp():
    """Tests for IndexedTableFormatter.setUp.

        >>> from schooltool.table.catalog import IndexedTableFormatter
        >>> from zope.publisher.browser import TestRequest
        >>> formatter = IndexedTableFormatter(None, TestRequest())

    Our IndexedTableFormatter can only work with index dicts and
    IndexedColumns, so before passing all the items to the
    constructor of the parent class we must process them:

        >>> formatter.indexItems = lambda list: [
        ...     {'id': "_id_%s" % item.lower(), 'catalog': "Catalog"}
        ...     for item in list]

    If we pass some items to the formatter now, they should be converted
    into index dicts:

        >>> formatter.setUp(items=["Pete", "John", "Bill"], ommit=["John"])
        >>> sorted(item['id'] for item in formatter._items)
        ['_id_bill', '_id_pete']

    """


def doctest_IndexedTableFormatter_render():
    """Tests for IndexedTableFormatter.setUp.

        >>> from schooltool.table.catalog import IndexedTableFormatter
        >>> from zope.publisher.browser import TestRequest

        >>> class ZCTableFormatterStub(object):
        ...     def __init__(self, context, request, items, **kw):
        ...         self.context = context
        ...         self.request = request
        ...         self.items = items
        ...         self.cssClasses = {}
        ...         self.args = dict(kw)
        ...
        ...     def __call__(self):
        ...         print 'Rendering', self.context
        ...         print 'Items:', self.items
        ...         print 'CSS classes:', sorted(self.cssClasses.items())
        ...         print 'Keyword arguments passed at creation:'
        ...         for k, v in sorted(self.args.items()):
        ...             print '%s:' % k, repr(v)

        >>> formatter = IndexedTableFormatter(
        ...     '<Context>', TestRequest())

     As IndexedTableFormatter works with index dicts as items, normal columns
     cannot render them (or use in any other way).  We adapt columns to
     IndexedColumns before passing to the rendering table formatter.

        >>> formatter.items = lambda: 'items'
        >>> formatter.ommit = lambda items, ommit: items + ' without ommitted'
        >>> formatter.setUp(
        ...     columns=['Column A', 'Column B'],
        ...     table_formatter=ZCTableFormatterStub)

        >>> from schooltool.table.interfaces import IIndexedColumn
        >>> provideAdapter(lambda column: 'Wrapped %s' % column,
        ...                adapts=[None],
        ...                provides=IIndexedColumn)

        >>> formatter.render()
        Rendering <Context>
        Items: items without ommitted
        CSS classes: [('table', 'data')]
        Keyword arguments passed at creation:
        batch_size: ...
        batch_start: ...
        columns: ['Wrapped Column A', 'Wrapped Column B']
        prefix: ...
        sort_on: ...

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF
                   | doctest.REPORT_ONLY_FIRST_FAILURE
                   | doctest.NORMALIZE_WHITESPACE)
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=optionflags))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
More columns for tables.
"""
import datetime

from zope.app.dependable.interfaces import IDependable
from zope.interface import implementer, implements, classImplements
from zope.i18n.interfaces.locales import ICollator
from zope.i18n import translate
from zope.intid.interfaces import IIntIds
from zope.interface import implementsOnly
from zope.component import adapter, queryMultiAdapter
from zope.component import queryUtility
from zope.security.proxy import removeSecurityProxy

import zc.table
import zc.table.interfaces
import zc.table.column

from schooltool.common import stupid_form_key, getResourceURL
from schooltool.table.interfaces import ICheckboxColumn
from schooltool.table.interfaces import IIndexedColumn


class CheckboxColumn(zc.table.column.Column):
    """A columns with a checkbox

    The name and id of the checkbox are composed of the prefix keyword
    argument and __name__ (or other value if form_id_builder is specified) of
    the item being displayed.
    """
    implements(ICheckboxColumn)

    def __init__(self, prefix, name=None, title=None,
                 isDisabled=None, id_getter=None,
                 value_getter=None):
        super(CheckboxColumn, self).__init__(name=name, title=title)
        if isDisabled:
            self.isDisabled = isDisabled
        self.prefix = prefix
        if id_getter is None:
            self.id_getter = stupid_form_key
        else:
            self.id_getter = id_getter
        self.value_getter = value_getter

    def isDisabled(self, item):
        return False

    def renderCell(self, item, formatter):
        if not self.isDisabled(item):
            checked = False
            if self.value_getter is not None:
                checked = bool(self.value_getter(item))
            form_id = ".".join(filter(None, [self.prefix, self.id_getter(item)]))
            return '<input type="checkbox" name="%s" id="%s"%s />' % (
                form_id, form_id,
                checked and ' checked="checked"' or '')
        else:
            return ''


class DependableCheckboxColumn(CheckboxColumn):
    """A column that displays a checkbox that is disabled if item has dependables.

    The name and id of the checkbox are composed of the prefix keyword
    argument and __name__ of the item being displayed.
    """

    def __init__(self, *args, **kw):
        kw = dict(kw)
        self.show_disabled = kw.pop('show_disabled', True)
        super(DependableCheckboxColumn, self).__init__(*args, **kw)

    def renderCell(self, item, formatter):
        form_id = ".".join(filter(None, [self.prefix, self.id_getter(item)]))
        if self.hasDependents(item):
            if self.show_disabled:
                return '<input type="checkbox" name="%s" id="%s" disabled="disabled" />' % (form_id, form_id)
            else:
                return ''
        else:
            checked = form_id in formatter.request and 'checked="checked"' or ''
            return '<input type="checkbox" name="%s" id="%s" %s/>' % (
                form_id, form_id, checked)

    def hasDependents(self, item):
        # We cannot adapt security-proxied objects to IDependable.  Unwrapping
        # is safe since we do not modify anything, and the information whether
        # an object can be deleted or not is not classified.
        unwrapped_context = removeSecurityProxy(item)
        dependable = IDependable(unwrapped_context, None)
        if dependable is None:
            return False
        else:
            return bool(dependable.dependents())


class DateColumn(zc.table.column.GetterColumn):
    """Table column that displays dates.

    Sortable even when None values are around.
    """

    def getSortKey(self, item, formatter):
        if self.getter(item, formatter) is None:
            return datetime.date.min
        else:
            return self.getter(item, formatter)

    def cell_formatter(self, maybe_date, item, formatter):
        view = queryMultiAdapter((maybe_date, formatter.request),
                                 name='mediumDate',
                                 default=lambda: '')
        return view()


class LocaleAwareGetterColumn(zc.table.column.GetterColumn):
    """Getter columnt that has locale aware sorting."""

    implements(zc.table.interfaces.ISortableColumn)

    def getSortKey(self, item, formatter):
        collator = ICollator(formatter.request.locale)
        s = self.getter(item, formatter)
        return s and collator.key(s)


class ImageInputColumn(zc.table.column.Column):

    def __init__(self, prefix, title=None, name=None,
                 alt=None, library=None, image=None, id_getter=None):
        super(ImageInputColumn, self).__init__(title=title, name=name)
        self.prefix = prefix
        self.alt = alt
        self.library = library
        self.image = image
        if id_getter is None:
            self.id_getter = stupid_form_key
        else:
            self.id_getter = id_getter

    def params(self, item, formatter):
        image_url = getResourceURL(self.library, self.image, formatter.request)
        if not image_url:
            return None
        form_id = ".".join(filter(None, [self.prefix, self.id_getter(item)]))
        return {
            'title': translate(self.title, context=formatter.request) or '',
            'alt': translate(self.alt, context=formatter.request) or '',
            'name': form_id,
            'src': image_url,
            }

    def renderCell(self, item, formatter):
        params = self.params(item, formatter)
        if not params:
            return ''
        return self.template() % params

    def template(self):
        return '\n'.join([
                '<button class="image" type="submit" name="%(name)s" title="%(title)s" value="1">',
                '<img src="%(src)s" alt="%(alt)s" />',
                '</button>'
                ])


class ImageInputValueColumn(ImageInputColumn):

    on_click = ""

    def __init__(self, *args, **kw):
        on_click = kw.pop('on_click', "")
        ImageInputColumn.__init__(self, *args, **kw)
        self.on_click = on_click
        self.form_id = ".".join(filter(None, [self.prefix, self.name]))

    def params(self, item, formatter):
        image_url = getResourceURL(self.library, self.image, formatter.request)
        if not image_url:
            return None
        value = self.id_getter(item)
        return {
            'title': translate(self.title, context=formatter.request) or '',
            'alt': translate(self.alt, context=formatter.request) or '',
            'name': self.form_id,
            'value': value,
            'src': image_url,
            'on_click': self.on_click,
            }

    def template(self):
        return '\n'.join([
                '<button class="image" type="submit" name="%(name)s"'
                ' title="%(title)s" value="%(value)s" onclick="%(on_click)s">',
                '<img src="%(src)s" alt="%(alt)s" />',
                '</button>'
                ])


class TableSubmitImageColumn(ImageInputColumn):

    def __init__(self, prefix, title=None, name=None, alt=None,
                 library='schooltool.skin.flourish', image=None,
                 container_id='', id_getter=None):
        super(TableSubmitImageColumn, self).__init__(
            prefix, title=title, name=name, alt=alt,
            library=library, image=image, id_getter=id_getter)
        self.form_id = ".".join(filter(None, [self.prefix, self.name]))
        self.container_id = container_id

    def on_click(self, formatter):
        return "return ST.table.on_form_submit('%s', this)" % self.container_id

    def params(self, item, formatter):
        image_url = getResourceURL(
            self.library, self.image, formatter.request)
        if not image_url:
            return None
        return {
            'title': translate(self.title, context=formatter.request) or '',
            'alt': translate(self.alt, context=formatter.request) or '',
            'name': self.form_id,
            'value': self.id_getter(item),
            'src': image_url,
            'script': self.on_click(formatter), # XXX: should encode this
            }

    def renderCell(self, item, formatter):
        params = self.params(item, formatter)
        if not params:
            return ''
        return self.template() % params

    def template(self):
        return ' '.join(s.strip() for s in [
                '<button class="image" type="submit" name="%(name)s"'
                '        title="%(title)s" value="%(value)s"',
                '        onclick="%(script)s">',
                '  <img src="%(src)s" alt="%(alt)s" />',
                '</button>'
                ])


class MultiStateColumn(zc.table.column.Column):

    header_column = None
    state_columns = None

    def __init__(self, states=(), header=None, state_getter=None):
        zc.table.column.Column.__init__(self)
        self.header_column = header
        self.state_columns = states
        if state_getter is None:
            state_getter = lambda item: 0
        self.getState = state_getter

    def renderHeader(self, formatter):
        return self.header_column.renderHeader(formatter)

    def renderCell(self, item, formatter):
        state = self.getState(item, formatter)
        column = self.state_columns[state]
        cell = column.renderCell(item, formatter)
        return cell


class IndexedGetterColumn(zc.table.column.GetterColumn):
    implements(IIndexedColumn, zc.table.interfaces.ISortableColumn)

    def __init__(self, **kwargs):
        self.index = kwargs.pop('index')
        super(IndexedGetterColumn, self).__init__(**kwargs)

    def _sort(self, items, formatter, start, stop, sorters, multiplier):
        if self.subsort and sorters:
            items = sorters[0](items, formatter, start, stop, sorters[1:])
        else:
            items = list(items) # don't mutate original
        getSortKey = self.getSortKey

        # Patch the SortableColum._sort to use both cmp and key for sorting.
        # This reduces usage of getSortKey drastically on large datasets.
        items.sort(
            cmp=lambda a, b: multiplier*cmp(a, b),
            key=lambda item: getSortKey(item, formatter))

        return items

    def renderCell(self, item, formatter):
        item = queryUtility(IIntIds).getObject(item['id'])
        return super(IndexedGetterColumn, self).renderCell(item, formatter)

    def getSortKey(self, item, formatter):
        id = item['id']
        index = item['catalog'][self.index]
        return index.documents_to_values[id]


class IndexedLocaleAwareGetterColumn(IndexedGetterColumn):

    _cached_collator = None

    def getSortKey(self, item, formatter):
        if not self._cached_collator:
            self._cached_collator = ICollator(formatter.request.locale)
        s = super(IndexedLocaleAwareGetterColumn, self).getSortKey(item, formatter)
        return s and self._cached_collator.key(s)


def makeIndexedColumn(mixins, column, *args, **kw):
    class_ = column.__class__
    new_class = type(
        '_indexed_%s' % class_.__name__,
        tuple(mixins) + (class_,),
        {'_non_indexed_column': column})
    classImplements(new_class, IIndexedColumn)
    new_column = super(class_, new_class).__new__(new_class, *args, **kw)
    new_column.__dict__.update(dict(column.__dict__))
    return new_column


def unindex(indexed_item):
    return queryUtility(IIntIds).getObject(indexed_item['id'])


class RenderUnindexingMixin(object):
    def renderCell(self, indexed_item, formatter):
        return super(RenderUnindexingMixin, self).renderCell(
            unindex(indexed_item), formatter)


class SortUnindexingMixin(object):
    def getSortKey(self, indexed_item, formatter):
        super(SortUnindexingMixin, self).getSortKey(
            unindex(indexed_item), formatter)


@adapter(zc.table.interfaces.IColumn)
@implementer(IIndexedColumn)
def getIndexedColumn(column):
    column = makeIndexedColumn(
        [RenderUnindexingMixin], column)
    return column


@adapter(zc.table.interfaces.ISortableColumn)
@implementer(IIndexedColumn)
def getIndexedSortableColumn(column):
    column = makeIndexedColumn(
        [RenderUnindexingMixin, SortUnindexingMixin], column)
    return column


class NoSortIndexedGetterColumn(IndexedGetterColumn):
    implementsOnly(IIndexedColumn)


class NoSortIndexedLocaleAwareGetterColumn(IndexedLocaleAwareGetterColumn):
    implementsOnly(IIndexedColumn)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool traverser metadirectives.
"""
from zope.security.zcml import Permission
from zope.schema import TextLine
from zope.interface import Interface
from zope.configuration.fields import GlobalObject, GlobalInterface


class ITraverser(Interface):
    for_ = GlobalInterface(
        title=u"The interface this traverser is for.",
        required=True,
        )

    type = GlobalInterface(
        title=u"Request type",
        required=True,
        )

    factory = GlobalObject(
        title=u"The pluggable traverser implementation.",
        required=True,
        )

    permission = Permission(
        title=u"Permission",
        description=u"The permission needed to use the view.",
        required=False,
        )

    provides = GlobalInterface(
        title=u"Interface the component provides",
        required=False,
        )


class IPluggableTraverser(ITraverser):

    factory = GlobalObject(
        title=u"The pluggable traverser implementation.",
        required=False,
        )


class ITraverserPluginBase(Interface):

    for_ = GlobalObject(
        title=u"The interface this plugin is for.",
        required=True,
        )

    layer = GlobalObject(
        title=u"The layer the plugin is declared for",
        required=False,
        )

    name = TextLine(
        title=u"The name the traverser will be traversing into.",
        required=False,
        )

    permission = Permission(
        title=u"Permission",
        required=False,
        )


class ITraverserPlugin(ITraverserPluginBase):
    """Traverser plugin zcml directive."""

    plugin = GlobalObject(
        title=u"The plugin that does the traversal.",
        required=True,
        )


class INamedTraverserPlugin(ITraverserPluginBase):
    """A simple safeguard against rogue generic traversers.

    Make traversal name mandatory.
    """
    name = TextLine(
        title=u"The name the traverser will be traversing into.",
        required=True,
        )


class INullTraverserPlugin(INamedTraverserPlugin):
    """Null traverser plugin zcml directive.

    The traverser returns the context.
    """


class IAttributeTraverserPlugin(INamedTraverserPlugin):
    """Attribute traverser plugin zcml directive.

    Traverses to an attribute of the context.
    """


class IAdapterTraverserPlugin(INamedTraverserPlugin):
    """Adapter traverser plugin zcml directive.

    Adapts context to the given interface.
    """

    adapter = GlobalObject(
        title=u"The interface this plugin will be adapting to.",
        required=True,
        )

    adapter_name = TextLine(
        title=u"Adapter name to use when adapting to given interface.",
        required=False,
        )



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Batch tests
"""
import unittest
import doctest

from zope.publisher.browser import TestRequest


class FormatterStub(object):
    def extra_url(self):
        return ""


def doctest_Batch_needsBatch():
    """Tests for needsBatch property

    Let's create a Batch

        >>> from schooltool.table.batch import Batch
        >>> formatter = FormatterStub()
        >>> formatter.request = TestRequest()
        >>> formatter._items = []
        >>> formatter.prefix = "persons"
        >>> batch = Batch(formatter)

     If there are no items - we don't need batching:

        >>> batch.needsBatch
        False

     If we have more items than the size of the page we will show
     batching controls:

        >>> batch.full_size = 29
        >>> batch.start = 0
        >>> batch.needsBatch
        True

     Now if we go to the next page, we should still see batching:

        >>> batch.start = 25
        >>> batch.needsBatch
        True

     Though if we only have as many items as the page size - we should
     see no batching:

        >>> batch.full_size = 25
        >>> batch.start = 0
        >>> batch.needsBatch
        False

    Unless we somehow paged less than the page size items forward:

        >>> batch.full_size = 25
        >>> batch.start = 10
        >>> batch.needsBatch
        True

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF
                   | doctest.REPORT_ONLY_FIRST_FAILURE
                   | doctest.NORMALIZE_WHITESPACE)
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(optionflags=optionflags))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for basic person views.
"""
import unittest
import doctest

from z3c.form.testing import TestRequest
from zope.app.testing import setup
from zope.app.testing.setup import setUpAnnotations
from zope.component import provideAdapter
from zope.interface import directlyProvides
from zope.interface import implements
from zope.schema.vocabulary import getVocabularyRegistry
from zope.traversing.interfaces import IContainmentRoot

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.group.interfaces import IGroupContainer
from schooltool.relationship.tests import setUpRelationships


def doctest_PersonContainerView():
    r"""Test for PersonContainerView

    Let's create some persons to delete from a person container:

        >>> from schooltool.basicperson.browser.person import \
        ...     BasicPersonContainerView
        >>> from schooltool.person.person import Person, PersonContainer
        >>> setup.setUpAnnotations()

        >>> personContainer = PersonContainer()
        >>> directlyProvides(personContainer, IContainmentRoot)

        >>> personContainer['pete'] = Person('pete', 'Pete Parrot')
        >>> personContainer['john'] = Person('john', 'Long John')
        >>> personContainer['frog'] = Person('frog', 'Frog Man')
        >>> personContainer['toad'] = Person('toad', 'Taodsworth')

        >>> request = TestRequest()
        >>> view = BasicPersonContainerView(personContainer, request)

    Our user is not trying to delete anything yet:

        >>> view.isDeletingHimself()
        False

    Lets log in:

        >>> from schooltool.app.security import Principal
        >>> principal = Principal('pete', 'Pete Parrot', personContainer['pete'])
        >>> request.setPrincipal(principal)

    Even if he is trying to delete someone who is not pete:

        >>> request.form = {'delete.frog': 'on',
        ...                 'delete.toad': 'on'}
        >>> view.isDeletingHimself()
        False

    But if he will try deleting himself - the method should return true:

        >>> request.form = {'delete.pete': 'on',
        ...                 'delete.toad': 'on'}
        >>> view.isDeletingHimself()
        True

    """


def doctest_PersonAddFormAdapter():
    """Tests for PersonAddFormAdapter

    PersonAddFormAdapter just wraps a person object but hides
    setPassword method unde a write only password property:

        >>> class PersonStub(object):
        ...     def setPassword(self, new_password):
        ...         print "Setting password to:", new_password

        >>> person = PersonStub()
        >>> person.name = "John"
        >>> person.last_name = "Johnson"

        >>> from schooltool.basicperson.browser.person import PersonAddFormAdapter
        >>> pa = PersonAddFormAdapter(person)
        >>> pa.password = "FooBar3"
        Setting password to: FooBar3

        >>> pa.password is None
        True

        >>> pa.name
        'John'

        >>> pa.surname = "Peterson"
        >>> person.surname
        'Peterson'

    """


class AppStub(dict):
    implements(ISchoolToolApplication)


def doctest_PersonAddView():
    r"""Test for PersonAddView

    Make sure we have the PersonFactory utility available:

        >>> from zope.component import provideUtility
        >>> from schooltool.basicperson.person import PersonFactoryUtility
        >>> from schooltool.person.interfaces import IPersonFactory
        >>> provideUtility(PersonFactoryUtility(), IPersonFactory)

    We need some setup to make traversal work in a unit test.

        >>> class FakeURL:
        ...     def __init__(self, context, request): pass
        ...     def __call__(self): return "http://127.0.0.1/frogpond/persons"
        ...
        >>> from schooltool.person.interfaces import IPersonContainer
        >>> from zope.traversing.browser.interfaces import IAbsoluteURL
        >>> provideAdapter(FakeURL, (IPersonContainer, TestRequest), IAbsoluteURL)

        >>> app = AppStub()
        >>> from zope.component import provideAdapter
        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> provideAdapter(lambda context: app,
        ...                adapts=[None],
        ...                provides=ISchoolToolApplication)

    Let's create a PersonContainer

        >>> from schooltool.person.person import PersonContainer
        >>> pc = PersonContainer()

    And a group container:

        >>> from schooltool.group.group import GroupContainer
        >>> gc = GroupContainer()
        >>> provideAdapter(lambda context: gc,
        ...                adapts=[ISchoolToolApplication],
        ...                provides=IGroupContainer)

        >>> from schooltool.basicperson.interfaces import IDemographicsFields
        >>> from schooltool.basicperson.demographics import DemographicsFields
        >>> df = DemographicsFields()
        >>> provideAdapter(lambda context: df,
        ...                adapts=[ISchoolToolApplication],
        ...                provides=IDemographicsFields)

    Now let's create a PersonAddView for the container

        >>> from schooltool.basicperson.browser.person import PersonAddFormAdapter
        >>> provideAdapter(PersonAddFormAdapter)
        >>> from schooltool.basicperson.browser.person import UsernameValidator
        >>> provideAdapter(UsernameValidator)
        >>> from schooltool.basicperson.browser.person import PasswordValidator
        >>> provideAdapter(PasswordValidator)

        >>> from schooltool.basicperson.browser.person import PersonAddView
        >>> view = PersonAddView(pc, TestRequest())
        >>> view.update()

    Let's try to add a user:

        >>> request = TestRequest(form={'form.widgets.first_name': u'John',
        ...                             'form.widgets.last_name': u'Doe',
        ...                             'form.widgets.username': u'jdoe',
        ...                             'form.widgets.password': u'secret',
        ...                             'form.widgets.confirm': u'secret',
        ...                             'form.buttons.add': 'Add'})
        >>> view = PersonAddView(pc, request)
        >>> view.update()
        >>> view.widgets.errors
        ()
        >>> 'jdoe' in pc
        True
        >>> person = pc['jdoe']
        >>> person.title
        u'John Doe'
        >>> person.username
        u'jdoe'
        >>> person.checkPassword('secret')
        True

    If we try to add a user with the same login, we get a nice error message:

        >>> request = TestRequest(form={'form.widgets.first_name': u'Another John',
        ...                             'form.widgets.last_name': u'Doe',
        ...                             'form.widgets.username': u'jdoe',
        ...                             'form.widgets.password': u'pass',
        ...                             'form.widgets.confirm': u'pass',
        ...                             'form.buttons.add': 'Add'})
        >>> view = PersonAddView(pc, request)
        >>> view.update()
        >>> for error in view.widgets.errors: print error.message
        This username is already in use

    Let's try to add user with different password and confirm password fields:

        >>> request = TestRequest(form={'form.widgets.first_name': u'Coo',
        ...                             'form.widgets.last_name': u'Guy',
        ...                             'form.widgets.username': u'coo',
        ...                             'form.widgets.password': u'secret',
        ...                             'form.widgets.confirm': u'plain',
        ...                             'form.buttons.add': 'Add'})
        >>> view = PersonAddView(pc, request)
        >>> view.update()
        >>> for error in view.widgets.errors: print error.message
        Passwords do not match
        >>> 'coo' in pc
        False

    We can select groups that the user should be in.  First, let's create a
    group:

        >>> from schooltool.group.group import Group
        >>> pov = IGroupContainer(app)['pov'] = Group('PoV')

    Now, let's create and render a view:

        >>> request = TestRequest(form={'form.widgets.first_name': u'Gintas',
        ...                             'form.widgets.last_name': u'Mil',
        ...                             'form.widgets.username': u'gintas',
        ...                             'form.widgets.password': u'denied',
        ...                             'form.widgets.confirm': u'denied',
        ...                             'form.widgets.group': ['pov'],
        ...                             'form.buttons.add': 'Add'})
        >>> view = PersonAddView(pc, request)
        >>> view.update()
        >>> view.widgets.errors
        ()

    Now the person belongs to the group that we have selected:

        >>> list(pc['gintas'].groups) == [pov]
        True

    We can cancel an action if we want to:

        >>> request = TestRequest(form={'form.buttons.cancel': 'Cancel'})
        >>> view = PersonAddView(pc, request)
        >>> view.update()
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/frogpond/persons'

    """


def doctest_PersonTerm():
    """Tests for PersonTerm.

    Person term is a title tokenized term that uses the title of a
    person as the title to be displayed:

        >>> class PersonStub(object):
        ...     def __init__(self, title):
        ...         self.title = title
        ...         self.__name__ = title.lower()

        >>> john = PersonStub('John')

        >>> from schooltool.basicperson.browser.person import PersonTerm
        >>> term = PersonTerm(john)
        >>> term.title
        'John'
        >>> term.token
        'john'
        >>> term.value
        <...test_person.PersonStub object at ...>

    """


def doctest_TermsBase():
    """Tests for TermsBase.

    Let's construct the TermsBase:

        >>> class TermStub(object):
        ...     def __init__(self, value):
        ...         self.value = value
        ...     def __repr__(self):
        ...         return "<TermStub %s>" % self.value

        >>> from schooltool.basicperson.browser.person import TermsBase
        >>> source = ["john"]
        >>> terms = TermsBase(source, None)

    If no term factory is set - NotImplementedError is raised:

        >>> terms.getTerm("john")
        Traceback (most recent call last):
        ...
        NotImplementedError: Term Factory must be provided by inheriting classes.

    If term factory is present it is used to construct the term from
    the given value:

        >>> terms.factory = TermStub

        >>> terms.getTerm("john")
        <TermStub john>

    If there is no such value in the source - we get a lookup error:

        >>> terms.getTerm("peter")
        Traceback (most recent call last):
        ...
        LookupError: peter

    """


def doctest_GroupTerm():
    """Tests for GroupTerm.

    Group term is a title tokenized term that uses the title of a
    group as the title to be displayed:

        >>> class GroupStub(object):
        ...     def __init__(self, title):
        ...         self.title = title
        ...         self.__name__ = title

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> class STAppStub(dict):
        ...     implements(ISchoolToolApplication)
        ...     def __init__(self, context):
        ...         self['groups'] = {'teachers': GroupStub('Teachers')}

        >>> provideAdapter(STAppStub, adapts=[None])
        >>> from schooltool.group.interfaces import IGroupContainer
        >>> provideAdapter(lambda app: app['groups'], adapts=[ISchoolToolApplication],
        ...                provides=IGroupContainer)

        >>> from schooltool.basicperson.browser.person import GroupTerm
        >>> term = GroupTerm(GroupStub("teachers"))
        >>> term.title
        'teachers'
        >>> term.token
        'teachers'
        >>> term.value
        <schooltool.basicperson.browser.tests.test_person.GroupStub object at ...>

    """


def setUp(test):
    setup.placelessSetUp()
    from z3c.form import testing
    testing.setupFormDefaults()
    vr = getVocabularyRegistry()
    from schooltool.basicperson.vocabularies import groupVocabularyFactory
    from schooltool.basicperson.vocabularies import advisorVocabularyFactory
    vr.register('schooltool.basicperson.group_vocabulary',
                groupVocabularyFactory())
    vr.register('schooltool.basicperson.advisor_vocabulary',
                advisorVocabularyFactory())
    setUpRelationships()
    setUpAnnotations()


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    return doctest.DocTestSuite(optionflags=optionflags,
                                setUp=setUp, tearDown=tearDown)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool timetable integration.
"""

from persistent.list import PersistentList
import zope.lifecycleevent
from zope.interface import implements
from zope.component import adapts

from schooltool.common import DateRange
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISubscriber
from schooltool.schoolyear.subscriber import EventAdapterSubscriber
from schooltool.term.term import getTermForDate, EmergencyDayEvent
from schooltool.timetable.interfaces import ITimetableContainer
from schooltool.timetable.interfaces import IScheduleExceptions
from schooltool.timetable.schedule import MeetingException


class EmergencyDayTimetableSubscriber(EventAdapterSubscriber):
    adapts(EmergencyDayEvent)
    implements(ISubscriber)

    def __call__(self):
        term = getTermForDate(self.event.date)
        if term is None:
            return
        old_date = self.event.date
        new_date = self.event.replacement_date
        schoolyear = ISchoolYear(term)
        timetables = ITimetableContainer(schoolyear)
        for timetable in timetables.values():
            if IScheduleExceptions.providedBy(timetable):
                modified = False
                scheduled = DateRange(timetable.first, timetable.last)
                meeting_exceptions = PersistentList()
                if old_date in scheduled:
                    meetings = list(timetable.iterMeetings(old_date))
                    for meeting in meetings:
                        meeting_exceptions.append(
                            MeetingException(
                                meeting.dtstart.replace(year=new_date.year,
                                                        month=new_date.month,
                                                        day=new_date.day),
                                meeting.duration,
                                period=meeting.period,
                                meeting_id=meeting.meeting_id))
                    timetable.exceptions[old_date] = PersistentList()
                    modified = True
                if new_date in scheduled:
                    timetable.exceptions[new_date] = meeting_exceptions
                    modified = True
                if modified:
                    zope.lifecycleevent.modified(timetable)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.person.person.
"""
import unittest
import doctest

from zope.interface.verify import verifyObject
from zope.app.testing import setup

from schooltool.person.person import PersonPasswordWriter
from schooltool.person.interfaces import IPasswordWriter

from schooltool.testing import setup as sbsetup


def doctest_PersonContainer():
    """Tests for PersonContainer

        >>> from schooltool.person.interfaces import IPersonContainer
        >>> from schooltool.person.person import PersonContainer
        >>> c = PersonContainer()
        >>> verifyObject(IPersonContainer, c)
        True

    PersonContainer uses the `username` attribute of persons as the key

        >>> from schooltool.person.person import Person
        >>> person = Person(username="itsme")
        >>> c['doesnotmatter'] = person
        >>> c['itsme'] is person
        True
        >>> c.get('doesnotmatter') is None
        True

    Adaptation (i.e. __conform__) is tested in doctest_SchoolToolApplication.
    """


def doctest_Person():
    r"""Tests for Person

        >>> from schooltool.person.interfaces import IPersonContained
        >>> from schooltool.person.person import Person
        >>> person = Person('person')
        >>> verifyObject(IPersonContained, person)
        True

    Persons initially have no password

        >>> person.hasPassword()
        False

    When a person has no password, he cannot log in

        >>> person.checkPassword('')
        False
        >>> person.checkPassword(None)
        False

    You can set the password

        >>> person.setPassword('secret')
        >>> person.hasPassword()
        True
        >>> person.checkPassword('secret')
        True
        >>> person.checkPassword('justguessing')
        False

    Note that the password is not stored in plain text and cannot be recovered

        >>> import pickle
        >>> 'secret' not in pickle.dumps(person)
        True

    You can lock out the user's accound by setting the password to None

        >>> person.setPassword(None)
        >>> person.hasPassword()
        False
        >>> person.checkPassword('')
        False
        >>> person.checkPassword(None)
        False

    Note that you can set the password to an empty string, although that is
    not a secure password

        >>> person.setPassword('')
        >>> person.hasPassword()
        True
        >>> person.checkPassword('')
        True
        >>> person.checkPassword(None)
        False

    It is probably not a very good idea to use non-ASCII characters in
    passwords, but you can do that

        >>> person.setPassword(u'\u1234')
        >>> person.checkPassword(u'\u1234')
        True

    Persons are comparable (by username):

        >>> person == person
        True
        >>> person == Person('person')
        True
        >>> person == 'foo'
        False

    Persons have a calendar:

        >>> setup.placelessSetUp()
        >>> setup.setUpAnnotations()
        >>> sbsetup.setUpCalendaring()

        >>> from schooltool.app.interfaces import ISchoolToolCalendar
        >>> calendar = ISchoolToolCalendar(person)
        >>> calendar.__name__
        'calendar'
        >>> calendar.__parent__ is person
        True
        >>> len(calendar)
        0

    Clean up:

        >>> setup.placelessTearDown()

    """


def doctest_PersonPreferences():
    r"""Tests for the Preferences adapter

        >>> from zope.app.testing import setup
        >>> setup.placelessSetUp()
        >>> setup.setUpAnnotations()
        >>> from schooltool.person.person import Person

        >>> person = Person('person')

    Note that the ``IHavePreferences`` interface is added in ZCML.

    Make sure the attribute stores the correct interface

        >>> from schooltool.person.interfaces import IPersonPreferences
        >>> from schooltool.person.preference import getPersonPreferences
        >>> prefs = getPersonPreferences(person)
        >>> verifyObject(IPersonPreferences, prefs)
        True

        >>> prefs
        <schooltool.person.preference.PersonPreferences object at 0x...>

        >>> prefs.cal_periods
        True

    Need to have prefs.__parent__ refer to the person it's attached to:

        >>> prefs.__parent__ == person
        True

    Called another time, getPersonPreferences() returns the same object:

        >>> getPersonPreferences(person) is prefs
        True

    Clean up:

        >>> setup.placelessTearDown()

    """


class TestPersonPasswordWriter(unittest.TestCase):

    def testSetPassword(self):
        from schooltool.person.person import Person
        person =  Person("Frog")
        passwordWriter = PersonPasswordWriter(person)
        passwordWriter.setPassword("gorf")
        self.assert_(person.checkPassword("gorf"))

    def testConformance(self):
        from schooltool.person.person import Person
        person =  Person("Frog")
        passwordWriter = PersonPasswordWriter(person)
        self.assert_(verifyObject(IPasswordWriter, passwordWriter))


def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(optionflags=doctest.ELLIPSIS),
        doctest.DocTestSuite('schooltool.person.person'),
        unittest.makeSuite(TestPersonPasswordWriter)
        ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2003 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Things common to the SchoolTool server and clients.
"""

import re
import locale
import datetime
import urllib
import HTMLParser

import zope.interface
import zope.component
from zope.publisher.interfaces import IApplicationRequest
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.security.management import queryInteraction
from zope.schema import Date
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.interface import Interface, implements

__metaclass__ = type


def parse_date(value):
    """Parse a ISO-8601 YYYY-MM-DD date value.

    Examples:

        >>> parse_date('2003-09-01')
        datetime.date(2003, 9, 1)
        >>> parse_date('20030901')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '20030901'
        >>> parse_date('2003-IX-01')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '2003-IX-01'
        >>> parse_date('2003-09-31')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '2003-09-31'
        >>> parse_date('2003-09-30-15-42')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '2003-09-30-15-42'

    """
    try:
        y, m, d = map(int, value.split('-'))
        return datetime.date(y, m, d)
    except ValueError:
        raise ValueError("Invalid date: %r" % value)


def parse_time(value):
    """Parse a ISO 8601 HH:MM time value.

    Examples:

        >>> parse_time('01:25')
        datetime.time(1, 25)
        >>> parse_time('9:15')
        datetime.time(9, 15)
        >>> parse_time('12:1')
        datetime.time(12, 1)
        >>> parse_time('00:00')
        datetime.time(0, 0)
        >>> parse_time('23:59')
        datetime.time(23, 59)
        >>> parse_time('24:00')
        Traceback (most recent call last):
          ...
        ValueError: Invalid time: '24:00'
        >>> parse_time('06:30PM')
        Traceback (most recent call last):
          ...
        ValueError: Invalid time: '06:30PM'

    """
    try:
        h, m = map(int, value.split(':'))
        return datetime.time(h, m)
    except ValueError:
        raise ValueError("Invalid time: %r" % value)


def parse_time_range(value, default_duration=None):
    """Parse a range of times (e.g. 9:45-14:20).

    This is an alternative implementation of time span parsing, taken from
    schooltool timetabling package.

    Example:

        >>> parse_time_range('9:45-14:20')
        (datetime.time(9, 45), datetime.timedelta(0, 16500))

        >>> parse_time_range('00:00-24:00')
        (datetime.time(0, 0), datetime.timedelta(1))

        >>> parse_time_range('10:00-10:00')
        (datetime.time(10, 0), datetime.timedelta(0))

    Time ranges may span midnight

        >>> parse_time_range('23:00-02:00')
        (datetime.time(23, 0), datetime.timedelta(0, 10800))

    When the default duration is specified, you may omit the second time

        >>> parse_time_range('23:00', 45)
        (datetime.time(23, 0), datetime.timedelta(0, 2700))

    Invalid values cause a ValueError

        >>> parse_time_range('something else')
        Traceback (most recent call last):
          ...
        ValueError: bad time range: something else

        >>> parse_time_range('9:00')
        Traceback (most recent call last):
          ...
        ValueError: duration is not specified

        >>> parse_time_range('9:00-9:75')
        Traceback (most recent call last):
          ...
        ValueError: minute must be in 0..59

        >>> parse_time_range('5:99-6:00')
        Traceback (most recent call last):
          ...
        ValueError: minute must be in 0..59

        >>> parse_time_range('14:00-24:01')
        Traceback (most recent call last):
          ...
        ValueError: hour must be in 0..23

    White space can be inserted between times

        >>> parse_time_range(' 9:45 - 14:20 ')
        (datetime.time(9, 45), datetime.timedelta(0, 16500))

    but not inside times

        >>> parse_time_range('9: 45-14:20')
        Traceback (most recent call last):
          ...
        ValueError: bad time range: 9: 45-14:20

    """
    m = re.match(r'\s*(\d+):(\d+)\s*(?:-\s*(\d+):(\d+)\s*)?$', value)
    if not m:
        raise ValueError('bad time range: %s' % value)
    h1, m1 = int(m.group(1)), int(m.group(2))
    if not m.group(3):
        if default_duration is None:
            raise ValueError('duration is not specified')
        duration = default_duration
    else:
        h2, m2 = int(m.group(3)), int(m.group(4))
        if (h2, m2) != (24, 0):   # 24:00 is expressly allowed here
            datetime.time(h2, m2) # validate the second time
        duration = (h2*60+m2) - (h1*60+m1)
        if duration < 0:
            duration += 1440
    return datetime.time(h1, m1), datetime.timedelta(minutes=duration)


def format_time_range(start, duration):
    """Format a range of times (e.g. 9:45-14:20).

    This is an alternative implementation of time span formatting, taken from
    schooltool timetabling package.

    Example:

        >>> format_time_range(datetime.time(9, 45),
        ...                   datetime.timedelta(0, 16500))
        '09:45-14:20'

        >>> format_time_range(datetime.time(0, 0), datetime.timedelta(1))
        '00:00-24:00'

        >>> format_time_range(datetime.time(10, 0), datetime.timedelta(0))
        '10:00-10:00'

        >>> format_time_range(datetime.time(23, 0),
        ...                   datetime.timedelta(0, 10800))
        '23:00-02:00'

    """
    end = (datetime.datetime.combine(datetime.date.today(), start) + duration)
    ends = end.strftime('%H:%M')
    if ends == '00:00' and duration == datetime.timedelta(1):
        return '00:00-24:00' # special case
    else:
        return '%s-%s' % (start.strftime('%H:%M'), ends)



def parse_datetime(s):
    """Parse a ISO 8601 date/time value.

    Only a small subset of ISO 8601 is accepted:

      YYYY-MM-DD HH:MM:SS
      YYYY-MM-DD HH:MM:SS.ssssss
      YYYY-MM-DDTHH:MM:SS
      YYYY-MM-DDTHH:MM:SS.ssssss

    Returns a datetime.datetime object without a time zone.

    Examples:

        >>> parse_datetime('2003-04-05 11:22:33.456789')
        datetime.datetime(2003, 4, 5, 11, 22, 33, 456789)

        >>> parse_datetime('2003-04-05 11:22:33.456')
        datetime.datetime(2003, 4, 5, 11, 22, 33, 456000)

        >>> parse_datetime('2003-04-05 11:22:33.45678999')
        datetime.datetime(2003, 4, 5, 11, 22, 33, 456789)

        >>> parse_datetime('01/02/03')
        Traceback (most recent call last):
          ...
        ValueError: Bad datetime: 01/02/03

    """
    m = re.match(r"(\d+)-(\d+)-(\d+)[ T](\d+):(\d+):(\d+)([.](\d+))?$", s)
    if not m:
        raise ValueError("Bad datetime: %s" % s)
    ssssss = m.groups()[7]
    if ssssss:
        ssssss = int((ssssss + "00000")[:6])
    else:
        ssssss = 0
    y, m, d, hh, mm, ss = map(int, m.groups()[:6])
    return datetime.datetime(y, m, d, hh, mm, ss, ssssss)


locale_charset = locale.getpreferredencoding()


def to_locale(us):
    r"""Convert a Unicode string to the current locale encoding.

    Example:

        >>> from schooltool import common
        >>> old_locale_charset = common.locale_charset

        >>> common.locale_charset = 'UTF-8'
        >>> to_locale(u'\u263B')
        '\xe2\x98\xbb'

        >>> common.locale_charset = 'ASCII'
        >>> to_locale(u'Unrepresentable: \u263B')
        'Unrepresentable: ?'

        >>> locale_charset = old_locale_charset

    """
    return us.encode(locale_charset, 'replace')


def from_locale(s):
    r"""Convert an 8-bit string in locale encoding to Unicode.

    Example:

        >>> from schooltool import common
        >>> old_locale_charset = common.locale_charset

        >>> from_locale('xyzzy')
        u'xyzzy'

        >>> common.locale_charset = 'UTF-8'
        >>> from_locale('\xe2\x98\xbb')
        u'\u263b'

        >>> locale_charset = old_locale_charset

    """
    return unicode(s, locale_charset)


def looks_like_a_uri(uri):
    r"""Check if the argument looks like a URI string.

    Refer to http://www.ietf.org/rfc/rfc2396.txt for details.
    We're only approximating to the spec.

    Some examples of valid URI strings:

        >>> looks_like_a_uri('http://foo/bar?baz#quux')
        True
        >>> looks_like_a_uri('HTTP://foo/bar?baz#quux')
        True
        >>> looks_like_a_uri('mailto:root')
        True

    These strings are all invalid URIs:

        >>> looks_like_a_uri('2HTTP://foo/bar?baz#quux')
        False
        >>> looks_like_a_uri('\nHTTP://foo/bar?baz#quux')
        False
        >>> looks_like_a_uri('mailto:postmaster ')
        False
        >>> looks_like_a_uri('mailto:postmaster text')
        False
        >>> looks_like_a_uri('nocolon')
        False
        >>> looks_like_a_uri(None)
        False

    """
    uri_re = re.compile(r"^[A-Za-z][A-Za-z0-9+-.]*:\S\S*$")
    return bool(uri and uri_re.match(uri) is not None)


def unquote_uri(uri):
    r"""Unquote a URI.

       >>> unquote_uri('/terms/%C5%BEiema')
       u'/terms/\u017eiema'

       >>> unquote_uri(u'/terms/%C5%BEiema')
       u'/terms/\u017eiema'

    """
    return urllib.unquote(str(uri)).decode('UTF-8')


def collect(fn):
    """Convert a generator to a function that returns a list of items.

        >>> @collect
        ... def something(x, y, z):
        ...     yield x
        ...     yield y + z

        >>> something(1, 2, 3)
        [1, 5]

    """
    def collector(*args, **kw):
        return list(fn(*args, **kw))
    collector.__name__ = fn.__name__
    collector.__doc__ = fn.__doc__
    collector.__dict__ = fn.__dict__
    collector.__module__ = fn.__module__
    return collector


class IDateRange(Interface):
    """A range of dates (inclusive).

    If r is an IDateRange, then the following invariant holds:
    r.first <= r.last

    Note that empty date ranges cannot be represented.
    """

    first = Date(
        title=u"The first day of the period of time covered.")

    last = Date(
        title=u"The last day of the period covered.")

    def __iter__():
        """Iterate over all dates in the range from the first to the last."""

    def __contains__(date):
        """Return True if the date is within the range, otherwise False.

        Raises a TypeError if date is not a datetime.date.
        """

    def __len__():
        """Return the number of dates covered by the range."""

    def overlaps(date_range):
        """Return whether this daterange overlaps with the other daterange."""

    def intersection(other_range):
        """Return intersection of date ranges (or None)."""



class DateRange(object):
    """A date range implementation using the standard datetime module.

    Date ranges are low-level components that represent a date
    span. They are mainly used to implement the date handling in
    terms:

      >>> january = DateRange(datetime.date(2003, 1, 1), datetime.date(2003, 1, 31))
      >>> IDateRange.providedBy(january)
      True

    You can use date ranges to check whether a certain date is within
    the range:

      >>> datetime.date(2002, 12, 31) in january
      False
      >>> datetime.date(2003, 2, 1) in january
      False
      >>> datetime.date(2003, 1, 1) in january
      True
      >>> datetime.date(2003, 1, 12) in january
      True
      >>> datetime.date(2003, 1, 31) in january
      True

    You can ask the date range for the amount of dates it includes.

      >>> days = list(january)
      >>> len(days)
      31
      >>> len(january)
      31

    As you can see, the boundary dates are inclusive. You can also
    iterate through all the dates in the date range.

      >>> days = DateRange(
      ...     datetime.date(2003, 1, 1), datetime.date(2003, 1, 2))
      >>> list(days)
      [datetime.date(2003, 1, 1), datetime.date(2003, 1, 2)]

      >>> days = DateRange(
      ...     datetime.date(2003, 1, 1), datetime.date(2003, 1, 1))
      >>> list(days)
      [datetime.date(2003, 1, 1)]

    If the beginning of the of the date range is later than the end, a
    value error is raised:

      >>> DateRange(datetime.date(2003, 1, 2),
      ...           datetime.date(2003, 1, 1)) # doctest: +NORMALIZE_WHITESPACE
      Traceback (most recent call last):
      ...
      ValueError: Last date datetime.date(2003, 1, 1) less than
                  first date datetime.date(2003, 1, 2)

    """
    implements(IDateRange)

    def __init__(self, first, last):
        self.first = first
        self.last = last
        if last < first:
            raise ValueError("Last date %r less than first date %r" %
                             (last, first))

    def __iter__(self):
        date = self.first
        while date <= self.last:
            yield date
            date += datetime.date.resolution

    def __len__(self):
        return (self.last - self.first).days + 1

    def __contains__(self, date):
        return self.first <= date <= self.last

    def overlaps(self, date_range):
        """Return whether this daterange overlaps with the other daterange.

           >>> from datetime import timedelta as td
           >>> def parse_inverval(interval):
           ...     first_date = datetime.date(2005, 1, 1)
           ...     start = len(interval) - len(interval.lstrip())
           ...     length = len(interval.strip())
           ...     end = start + length - 1
           ...     return DateRange(first_date + td(days=start),
           ...                      first_date + td(days=end))

           >>> interval = parse_inverval(' [===]  ')
           >>> print interval.first, interval.last
           2005-01-02 2005-01-06

           >>> interval = parse_inverval('  [===]   ')
           >>> print interval.first, interval.last
           2005-01-03 2005-01-07

           >>> def overlap(interval1, interval2):
           ...     i1 = parse_inverval(interval1)
           ...     i2 = parse_inverval(interval2)
           ...     return i1.overlaps(i2)

           >>> overlap(' [===]                     ',
           ...         '       [===========]       ')
           False

           >>> overlap(' [====]                    ',
           ...         '       [===========]       ')
           False

           >>> overlap(' [=====]                   ',
           ...         '       [===========]       ')
           True

           >>> overlap(' [=======]                 ',
           ...         '       [===========]       ')
           True

           >>> overlap('       [===========]       ',
           ...         '       [===========]       ')
           True

           >>> overlap('        [=========]        ',
           ...         '       [===========]       ')
           True

           >>> overlap('                [========] ',
           ...         '       [===========]       ')
           True

           >>> overlap('                   [=====] ',
           ...         '       [===========]       ')
           True

           >>> overlap('                    [====] ',
           ...         '       [===========]       ')
           False

           >>> overlap('                     [===] ',
           ...         '       [===========]       ')
           False

           >>> overlap('     [===============]     ',
           ...         '       [===========]       ')
           True

        """
        return self.last >= date_range.first and self.first <= date_range.last

    def intersection(self, other_range):
        """Return whether this daterange overlaps with the other daterange.

           >>> dr = DateRange(datetime.date(2005, 2, 10),
           ...                datetime.date(2005, 2, 15))

           >>> dr_before = DateRange(datetime.date(2005, 2, 1),
           ...                       datetime.date(2005, 2, 9))

           >>> print dr.intersection(dr_before)
           None

           >>> dr_after = DateRange(datetime.date(2005, 2, 16),
           ...                      datetime.date(2005, 2, 19))

           >>> print dr.intersection(dr_after)
           None

           >>> intersection = dr.intersection(DateRange(
           ...     datetime.date(2005, 2, 1), datetime.date(2005, 2, 10)))
           >>> print intersection.first, intersection.last
           2005-02-10 2005-02-10

           >>> intersection = dr.intersection(DateRange(
           ...     datetime.date(2005, 2, 13), datetime.date(2005, 2, 25)))
           >>> print intersection.first, intersection.last
           2005-02-13 2005-02-15

           >>> intersection = dr.intersection(DateRange(
           ...     datetime.date(2005, 2, 15), datetime.date(2005, 2, 25)))
           >>> print intersection.first, intersection.last
           2005-02-15 2005-02-15

           >>> intersection = dr.intersection(DateRange(
           ...     datetime.date(2005, 2, 13), datetime.date(2005, 2, 14)))
           >>> print intersection.first, intersection.last
           2005-02-13 2005-02-14

        """
        if (other_range.last < self.first or
            other_range.first > self.last):
            return None
        return DateRange(max(self.first, other_range.first),
                         min(self.last, other_range.last))


_version = None

def get_version():
    global _version
    if _version is not None:
        return _version
    import pkg_resources
    _version = pkg_resources.get_distribution('schooltool').version
    return _version


def get_all_versions():
    import pkg_resources
    versions = []
    versions.append(pkg_resources.get_distribution('schooltool'))
    for entry in pkg_resources.iter_entry_points('z3c.autoinclude.plugin'):
        if (entry.name == 'target' and
            entry.module_name == 'schooltool'):
            versions.append(entry.dist)
    for entry in pkg_resources.iter_entry_points('schooltool.plugin_configuration'):
        if entry.dist not in versions:
            versions.append(entry.dist)
    return sorted(versions, key=lambda dist: dist.project_name)


def getRequestFromInteraction(request_type=IApplicationRequest):
    """Extract the browser request from the current interaction.

    Returns None when there is no interaction, or when the interaction has no
    participations that provide request_type.
    """
    interaction = queryInteraction()
    if interaction is not None:
        for participation in interaction.participations:
            if request_type.providedBy(participation):
                return participation
    return None


from zope.i18nmessageid import Message, MessageFactory
SchoolToolMessage = MessageFactory("schooltool")

def format_message(message, mapping=None):
    """Add mapping to a zope.i18nmessageid.Message."""
    assert isinstance(message, Message)
    return message.__class__(message, mapping=mapping)


def stupid_form_key(item):
    """Simple form key that uses item's __name__.
    Very unsafe: __name__ is expected to be a unicode string that contains
    who knows what.
    """
    return item.__name__


def simple_form_key(item):
    """Unique form key for items contained within a single container."""
    name = getattr(item, '__name__', None)
    if name is None:
        return None
    key = unicode(name).encode('punycode')
    key = urllib.quote(key)
    return key


class IResourceURIGetter(zope.interface.Interface):

    def __call__(library_name, resource_name):
        """Return resource from the library."""


def getResourceURL(library_name, resource_name, request):
    getter = zope.component.queryAdapter(
        request, IResourceURIGetter,
        name=library_name, default=None)
    if getter is None:
        getter = zope.component.queryAdapter(
            request, IResourceURIGetter,
            default=None)
    if getter is None:
        return None
    resource = getter(library_name, resource_name)
    return resource


class CommonResourceURL(object):
    zope.component.adapts(IBrowserRequest)
    zope.interface.implements(IResourceURIGetter)

    def __init__(self, request):
        self.request = request

    def __call__(self, library_name, resource_name):
        if not resource_name:
            return None
        if library_name is not None:
            library = zope.component.queryAdapter(self.request, name=library_name)
            resource = library.get(resource_name)
        else:
            resource = zope.component.queryAdapter(self.request, name=resource_name)
        if resource is None:
            return None
        return absoluteURL(resource, self.request)


def data_uri(payload, mime=None):
    payload = payload.encode('base64').replace('\n','')
    result = 'data:'
    if mime:
        result += mime + ';'
    result = result + 'base64,' + payload
    return result


class HTMLToText(HTMLParser.HTMLParser):

    def __init__(self):
        self.reset()
        self.text_lines = []

    def handle_data(self, data):
        self.text_lines.append(data)

    def get_data(self):
        return ''.join(self.text_lines)


# Launchpad projects assigned for packages
_package_projects = {}

def find_launchpad_project(traceback, default="schooltool"):
    global _package_projects
    project = default
    while traceback is not None:
        frame = traceback.tb_frame
        package = frame.f_globals.get('__package__')
        while package:
            if package in _package_projects:
                project = _package_projects[package]
                break
            package = package.rpartition('.')[0]
        traceback = traceback.tb_next
    return project


def register_lauchpad_project(package, project):
    global _package_projects
    _package_projects[package] = project
